#
# Copyright 2018-2023 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

from abc import ABC
from abc import abstractmethod
from enum import Enum
from importlib import import_module
import json
import re
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import TYPE_CHECKING
from typing import Union

# Prevent a circular reference by importing RuntimePipelineProcessor only during type-checking
if TYPE_CHECKING:
    from elyra.pipeline.processor import RuntimePipelineProcessor

from elyra.pipeline.pipeline_constants import DISABLE_NODE_CACHING
from elyra.pipeline.pipeline_constants import ENV_VARIABLES
from elyra.pipeline.pipeline_constants import KUBERNETES_POD_ANNOTATIONS
from elyra.pipeline.pipeline_constants import KUBERNETES_POD_LABELS
from elyra.pipeline.pipeline_constants import KUBERNETES_SECRETS
from elyra.pipeline.pipeline_constants import KUBERNETES_SHARED_MEM_SIZE
from elyra.pipeline.pipeline_constants import KUBERNETES_TOLERATIONS
from elyra.pipeline.pipeline_constants import MOUNTED_VOLUMES
from elyra.pipeline.pipeline_constants import PIPELINE_PARAMETERS
from elyra.pipeline.runtime_type import RuntimeProcessorType
from elyra.util.kubernetes import is_valid_annotation_key
from elyra.util.kubernetes import is_valid_annotation_value
from elyra.util.kubernetes import is_valid_kubernetes_key
from elyra.util.kubernetes import is_valid_kubernetes_resource_name
from elyra.util.kubernetes import is_valid_label_key
from elyra.util.kubernetes import is_valid_label_value


class PropertyInputType:
    """
    An object representing a single allowed input type for a PropertyAttribute
    object with optional custom default value, placeholder, and enum.
    """

    _python_defined_types = {
        "str": {"json_type": "string", "default_value": ""},
        "bool": {"json_type": "boolean", "default_value": False, "placeholder": " "},
        "int": {"json_type": "integer"},
        "float": {"json_type": "number"},
        "list": {"json_type": "string", "default_value": "[]"},
        "dict": {"json_type": "string", "default_value": "{}"},
    }

    def __init__(
        self,
        base_type: str,
        default_value: Optional[Any] = None,
        placeholder: Optional[Any] = None,
        enum: Optional[List[Any]] = None,
        **kwargs,
    ):
        # Get type info from passed in type dict, if available; default to standard python types
        allowed_input_types = kwargs.get("runtime_defined_types")
        if not allowed_input_types:
            allowed_input_types = self._python_defined_types

        if base_type not in allowed_input_types:
            raise ValueError(
                f"Invalid property type '{base_type}': valid types are {list(allowed_input_types.keys())}."
            )
        type_defaults = allowed_input_types[base_type]

        # Set types needed for schema
        self.base_type = base_type
        self.json_data_type = type_defaults.get("json_type")

        # Set extra schema details
        self.default_value = default_value if default_value is not None else type_defaults.get("default_value")
        self.placeholder = placeholder if placeholder is not None else type_defaults.get("placeholder")
        self.enum = enum
        self.type_title = type_defaults.get("type_title")


class PropertyAttribute:
    """
    An attribute of an ElyraProperty instance that provides the means to construct the
    schema for a property and contains information for processing property instances.
    """

    def __init__(
        self,
        attribute_id: str,
        description: Optional[str] = None,
        display_name: Optional[str] = None,
        allowed_input_types: Optional[List[PropertyInputType]] = None,
        hidden: Optional[bool] = False,
        required: Optional[bool] = False,
        pattern: Optional[str] = None,
    ):
        """
        :param attribute_id: a shorthand id for this attribute, e.g. "env_var"
        :param description: a description of this attribute
        :param display_name: the display name for this attribute
        :param allowed_input_types: a list of accepted PropertyInputTypes for this property
        :param hidden: whether this attribute should be hidden in the UI, preventing users from entering a value
        :param required: whether a value for this attribute is required
        :param pattern: a regex validation pattern to use for values given for this attribute
        """
        if not attribute_id:
            raise ValueError("Invalid Elyra property attribute: missing attribute_id.")

        self.id = attribute_id
        self.description = description
        self.title = display_name
        self.allowed_input_types = allowed_input_types or []
        self.default_type = self.allowed_input_types[0] if self.allowed_input_types else None
        self.hidden = hidden
        self.required = required
        self.pattern = pattern


class ListItemPropertyAttribute(PropertyAttribute):
    """
    An attribute of an ElyraPropertyListItem instance that provides the means to construct the
    schema for a property and contains information for processing property instances.
    """

    def __init__(
        self,
        attribute_id: str,
        description: Optional[str] = None,
        display_name: Optional[str] = None,
        allowed_input_types: Optional[List[PropertyInputType]] = None,
        hidden: Optional[bool] = False,
        required: Optional[bool] = False,
        pattern: Optional[str] = None,
        use_in_key: Optional[bool] = True,
    ):
        """
        :param use_in_key: whether this attribute should be used when constructing a
            key for an instance that will be used to de-duplicate list items
        """
        super().__init__(attribute_id, description, display_name, allowed_input_types, hidden, required, pattern)
        self.use_in_key = use_in_key


class ElyraProperty(ABC):
    """A component property that is defined and processed by Elyra"""

    applies_to_generic: bool  # True if the property applies to generic components
    applies_to_custom: bool  # True if the property applies to custom components

    property_id: str
    property_display_name: str
    property_description: str
    property_attributes: List[PropertyAttribute] = []
    _json_data_type: str = None
    _required: bool = False

    _subclass_property_map: Dict[str, type] = {}

    @classmethod
    def all_subclasses(cls):
        """Get all nested subclasses for a class."""
        return set(cls.__subclasses__()).union([s for c in cls.__subclasses__() for s in c.all_subclasses()])

    @classmethod
    def build_property_map(cls) -> None:
        """Build the map of property subclasses."""
        cls._subclass_property_map = {sc.property_id: sc for sc in cls.all_subclasses() if hasattr(sc, "property_id")}

    @classmethod
    def get_class_for_property(cls, prop_id) -> type | None:
        """Returns the ElyraProperty subclass corresponding to the given property id."""
        if not cls._subclass_property_map:
            cls.build_property_map()
        return cls._subclass_property_map.get(prop_id)

    @classmethod
    def subclass_exists_for_property(cls, prop_id: str) -> bool:
        """
        Returns a boolean indicating whether a corresponding ElyraProperty subclass
        exists for the given property id.
        """
        return cls.get_class_for_property(prop_id) is not None

    @classmethod
    def get_single_instance(cls, value: Optional[Dict[str, Any]] = None) -> ElyraProperty | None:
        """Unpack values from dictionary object and instantiate a class instance."""
        if isinstance(value, ElyraProperty):
            return value  # value is already a single instance, no further action required

        if not isinstance(value, dict):
            value = {}
        params = {attr.id: cls.strip_if_string(value.get(attr.id)) for attr in cls.property_attributes}
        instance = getattr(import_module(cls.__module__), cls.__name__)(**params)
        return None if instance.should_discard() else instance

    @classmethod
    def create_instance(cls, prop_id: str, value: Optional[Any]) -> ElyraProperty | ElyraPropertyList | None:
        """Create an instance of a class with the given property id using the user-entered values."""
        sc = cls.get_class_for_property(prop_id)
        if sc is None:
            return None

        if issubclass(sc, ElyraPropertyListItem):
            if not isinstance(value, list):
                return None
            instances = [sc.get_single_instance(obj) for obj in value]  # create instance for each object
            return ElyraPropertyList(instances).deduplicate()  # convert to ElyraPropertyList and de-dupe
        elif issubclass(sc, ElyraProperty):
            return sc.get_single_instance(value)
        return None

    @classmethod
    def get_classes_for_component_type(
        cls, component_type: str, runtime_type: Optional[RuntimeProcessorType] = None
    ) -> Set[type]:
        """
        Retrieve property subclasses that apply to the given component type
        (e.g., custom or generic) and to the given runtime type.
        """
        from elyra.pipeline.processor import PipelineProcessorManager  # placed here to avoid circular reference

        processor_props = set()
        for processor in PipelineProcessorManager.instance().get_all_processors():
            props = getattr(processor, "supported_properties", set())
            if not runtime_type or runtime_type == RuntimeProcessorType.LOCAL or processor.type == runtime_type:
                processor_props.update(props)

        all_subclasses = set()
        for sc in cls.all_subclasses():
            sc_id = getattr(sc, "property_id", "")
            if sc_id in processor_props and getattr(sc, f"applies_to_{component_type}", False):
                all_subclasses.add(sc)

        return all_subclasses

    @classmethod
    def get_schema(cls) -> Dict[str, Any]:
        """
        Build the JSON schema for an Elyra-owned component property using the attributes
        defined in the human-readable list of PropertyAttribute object for each class.
        """
        class_description = re.sub(" +", " ", cls.property_description.replace("\n", " ")).strip()
        schema = {"title": cls.property_display_name, "description": class_description, "type": cls._json_data_type}
        if cls._json_data_type is not None:  # property is a scalar value  TODO deprecate when able
            return schema

        properties, uihints, required_list = {}, {}, []
        for attr in cls.property_attributes:
            if attr.hidden:
                continue
            if attr.required:
                required_list.append(attr.id)

            properties[attr.id] = {"title": attr.title or attr.id}
            if attr.description:
                properties[attr.id]["description"] = attr.description
            if attr.pattern:
                properties[attr.id]["pattern"] = attr.pattern

            if len(attr.allowed_input_types) == 1:
                allowed_type: PropertyInputType = attr.allowed_input_types[0]

                properties[attr.id]["type"] = allowed_type.json_data_type
                if allowed_type.default_value is not None:
                    properties[attr.id]["default"] = allowed_type.default_value
                if allowed_type.enum:
                    properties[attr.id]["enum"] = allowed_type.enum
                if allowed_type.placeholder is not None:
                    uihints[attr.id] = {"ui:placeholder": allowed_type.placeholder}

            elif len(attr.allowed_input_types) > 1:
                properties[attr.id]["title"] = "Type"
                properties[attr.id]["description"] = "The type of this parameter"

                # Set default type to be the first one defined in the property_attributes list
                properties[attr.id]["default"] = {
                    "type": attr.allowed_input_types[0].base_type,
                    "value": attr.allowed_input_types[0].default_value,
                }

                # Render a oneOf block and loop through allowed types to fill it
                properties[attr.id]["oneOf"] = []
                for allowed_type in sorted(attr.allowed_input_types, key=lambda t: t.base_type):
                    schema_obj = {
                        "type": "object",
                        "title": allowed_type.type_title or allowed_type.base_type,
                        "properties": {
                            "type": {"type": "string", "default": allowed_type.type_title or allowed_type.base_type},
                            "value": {
                                "title": attr.title,
                                "description": attr.description,
                                "type": allowed_type.json_data_type,
                            },
                        },
                        "uihints": {"type": {"ui:widget": "hidden"}},
                    }
                    if allowed_type.default_value is not None:
                        schema_obj["properties"]["value"]["default"] = allowed_type.default_value
                    if allowed_type.enum:
                        schema_obj["properties"]["value"]["enum"] = allowed_type.enum
                    if allowed_type.placeholder is not None:
                        schema_obj["uihints"]["value"] = {"ui:placeholder": allowed_type.placeholder}

                    properties[attr.id]["oneOf"].append(schema_obj)

        if issubclass(cls, ElyraPropertyListItem):
            items = {"type": "object", "properties": properties, "required": required_list}
            schema.update({"type": "array", "default": [], "items": items, "uihints": {"items": uihints}})
        else:
            schema.update({"type": "object", "properties": properties, "required": required_list, "uihints": uihints})

        return schema

    @staticmethod
    def strip_if_string(var: Any) -> Any:
        """Strip surrounding whitespace from variable if it is a string"""
        return var.strip() if isinstance(var, str) else var

    def should_discard(self) -> bool:
        """
        Returns a boolean indicating whether an instance should be silently discarded on
        the basis of its attribute values. A discarded instance will not be validated or
        processed.

        Override this method if there are any constraints that dictate that this instance
        should not be processed.
        """
        return False

    @abstractmethod
    def get_value_for_display(self) -> Dict[str, Any]:
        """
        Get a representation of the instance to display in UI error messages.
        Should be implemented in any subclass that has validation criteria.
        """
        pass

    @abstractmethod
    def get_all_validation_errors(self) -> List[str]:
        """Perform custom validation on an instance."""
        pass

    @abstractmethod
    def add_to_execution_object(self, runtime_processor: RuntimePipelineProcessor, execution_object: Any, **kwargs):
        """
        Add a property instance to the execution object for the given runtime processor.
        Calls the runtime processor's implementation of add_{property_type}, e.g.
        runtime_processor.add_kubernetes_secret(self, execution_object, **kwargs).
        """
        pass


class DisableNodeCaching(ElyraProperty):
    """An ElyraProperty representing node cache preference"""

    applies_to_generic = False
    applies_to_custom = True

    property_id = DISABLE_NODE_CACHING
    property_display_name = "Disable node caching"
    property_description = "Disable caching to force node re-execution in the target runtime environment."
    _json_data_type = "string"

    def __init__(self, selection: Union[str, bool], **kwargs):
        self.selection = None
        if selection in ["True", "true", True]:
            self.selection = True
        elif selection in ["False", "false", False]:
            self.selection = False

    @classmethod
    def get_single_instance(cls, value: Optional[Any] = None) -> ElyraProperty | None:
        if isinstance(value, ElyraProperty):
            return value  # value is already a single instance, no further action required
        instance = DisableNodeCaching(selection=value)
        return None if instance.should_discard() else instance

    @classmethod
    def get_schema(cls) -> Dict[str, Any]:
        """Build the JSON schema for this property"""
        schema = super().get_schema()
        schema["enum"] = ["True", "False"]
        schema["uihints"] = {"ui:placeholder": "Use runtime default"}
        return schema

    def get_value_for_display(self) -> Dict[str, Any]:
        return self.selection

    def get_all_validation_errors(self) -> List[str]:
        return []

    def add_to_execution_object(self, runtime_processor: RuntimePipelineProcessor, execution_object: Any, **kwargs):
        """Add DisableNodeCaching info to the execution object for the given runtime processor"""
        runtime_processor.add_disable_node_caching(instance=self, execution_object=execution_object, **kwargs)

    def should_discard(self) -> bool:
        """Ignore this DisableNodeCaching instance if neither True nor False was selected."""
        return self.selection is None


class CustomSharedMemorySize(ElyraProperty):
    """An ElyraProperty representing shared memory size for a node."""

    applies_to_generic = True  # custom shared mem size applies to generic components
    applies_to_custom = True  # custom shared mem size applies to custom components

    property_id = KUBERNETES_SHARED_MEM_SIZE
    property_display_name = "Shared Memory Size"
    property_description = """Configure a custom shared memory size in
    gigabytes (10^9 bytes) for the pod that executes a node. A custom
    value is assigned if the size property value is a number greater than zero."""
    property_attributes = [
        PropertyAttribute(
            attribute_id="size",
            display_name="Memory Size (GB)",
            allowed_input_types=[PropertyInputType(base_type="int", placeholder=0)],
            hidden=False,
            required=False,
        ),
        PropertyAttribute(
            attribute_id="units",
            display_name="Units",
            allowed_input_types=[PropertyInputType(base_type="str")],
            hidden=True,
            required=False,
        ),
    ]

    default_units = "G"

    def __init__(self, size: str, units: str, **kwargs):
        self.size = size
        self.units = units or CustomSharedMemorySize.default_units

    @classmethod
    def get_schema(cls) -> Dict[str, Any]:
        """Build the JSON schema for an Elyra-owned component property"""
        schema = super().get_schema()
        schema["properties"]["size"]["minimum"] = 0
        return schema

    def to_dict(self) -> Dict[str, Any]:
        """Convert instance to a dict with relevant class attributes."""
        dict_repr = {attr.id: getattr(self, attr.id, None) for attr in self.property_attributes}
        return dict_repr

    def get_value_for_display(self) -> Dict[str, Any]:
        """Get a representation of the instance to display in UI error messages."""
        return self.to_dict()

    def get_all_validation_errors(self) -> List[str]:
        """Validate this instance. If the size attribute is set and not zero it
        must be a positive floating point number. The units attribute must be 'G'."""
        validation_errors = []
        # verify custom size
        try:
            if self.size:
                size = float(self.size)
                if size < 0:
                    raise ValueError()
        except ValueError:
            validation_errors.append(f"Shared memory size '{self.size}' must be a positive number.")
        # verify units
        if self.units not in ["G"]:
            validation_errors.append(f"Shared memory size units '{self.units}' must be 'G'.")
        return validation_errors

    def add_to_execution_object(self, runtime_processor: RuntimePipelineProcessor, execution_object: Any, **kwargs):
        """Add CustomSharedMemorySize instance to the execution object for the given runtime processor"""
        runtime_processor.add_custom_shared_memory_size(instance=self, execution_object=execution_object, **kwargs)

    def should_discard(self) -> bool:
        """Ignore this CustomSharedMemorySize instance if no custom size is specified."""
        return not self.size


class ElyraPropertyListItem(ElyraProperty):
    """
    An Elyra-owned property that is meant to be a member of an ElyraOwnedPropertyList.
    """

    property_attributes: List[ListItemPropertyAttribute] = []

    def to_dict(self) -> Dict[str, Any]:
        """Convert instance to a dict with relevant class attributes."""
        dict_repr = {attr.id: getattr(self, attr.id, None) for attr in self.property_attributes}
        return dict_repr

    def get_key_for_dict_entry(self) -> str:
        """
        Given the attribute names in the 'key' property, construct a key
        based on the attribute values of the instance.
        """
        prop_key = ""
        keys = [attr.id for attr in self.property_attributes if attr.use_in_key]
        for key_attr in keys:
            key_part = getattr(self, key_attr)
            if key_part:
                prop_key += f"{key_part}:" if key_attr != keys[-1] else key_part
        return prop_key

    def get_value_for_dict_entry(self) -> Union[str, Dict[str, Any]]:
        """Returns the value to be used when constructing a dict from a list of ElyraPropertyListItem."""
        return self.to_dict()

    def get_value_for_display(self) -> Dict[str, Any]:
        """Get a representation of the instance to display in UI error messages."""
        dict_repr = self.to_dict()
        for attr in self.property_attributes:
            if attr.hidden:
                dict_repr.pop(attr.id)
        return dict_repr


class EnvironmentVariable(ElyraPropertyListItem):
    """An ElyraProperty representing a single Environment Variable"""

    applies_to_generic = True
    applies_to_custom = False

    property_id = ENV_VARIABLES
    property_display_name = "Environment Variables"
    property_description = "Environment variables to be set on the execution environment."
    property_attributes = [
        ListItemPropertyAttribute(
            attribute_id="env_var",
            display_name="Environment Variable",
            allowed_input_types=[PropertyInputType(base_type="str", placeholder="ENV_VAR")],
            hidden=False,
            required=True,
            use_in_key=True,
        ),
        ListItemPropertyAttribute(
            attribute_id="value",
            display_name="Value",
            allowed_input_types=[PropertyInputType(base_type="str", placeholder="value")],
            hidden=False,
            required=False,
            use_in_key=False,
        ),
    ]

    def __init__(self, env_var, value, **kwargs):
        self.env_var = env_var
        self.value = value

    @classmethod
    def get_schema(cls) -> Dict[str, Any]:
        """Build the JSON schema for an Elyra-owned component property"""
        schema = super().get_schema()
        schema["uihints"].update({"canRefresh": True})
        return schema

    def get_value_for_dict_entry(self) -> Union[str, Dict[str, Any]]:
        """
        Returns the value to be used when constructing a dict from a list of ElyraPropertyListItem.
        A EnvironmentVariable dict entry will be of the form {self.env_var: self.value}
        """
        return self.value

    def should_discard(self) -> bool:
        """If a value is not specified, this EnvironmentVariable instance should be silently ignored."""
        return not self.value

    def get_all_validation_errors(self) -> List[str]:
        """Perform custom validation on an instance."""
        validation_errors = []
        if not self.env_var:
            validation_errors.append("Required environment variable was not specified.")
        elif " " in self.env_var:
            validation_errors.append(f"Environment variable '{self.env_var}' includes invalid space character(s).")

        return validation_errors

    def add_to_execution_object(self, runtime_processor: RuntimePipelineProcessor, execution_object: Any, **kwargs):
        """Add EnvironmentVariable instance to the execution object for the given runtime processor"""
        runtime_processor.add_env_var(instance=self, execution_object=execution_object, **kwargs)


class KubernetesSecret(ElyraPropertyListItem):
    """An ElyraProperty representing a single Kubernetes secret"""

    applies_to_generic = True
    applies_to_custom = False

    property_id = KUBERNETES_SECRETS
    property_display_name = "Kubernetes Secrets"
    property_description = """Kubernetes secrets to make available as environment
    variables to this node. The secret name and key given must be present in the
    Kubernetes namespace where the node is executed or this node will not run."""
    property_attributes = [
        ListItemPropertyAttribute(
            attribute_id="env_var",
            display_name="Environment Variable",
            allowed_input_types=[PropertyInputType(base_type="str", placeholder="ENV_VAR")],
            hidden=False,
            required=True,
            use_in_key=True,
        ),
        ListItemPropertyAttribute(
            attribute_id="name",
            display_name="Secret Name",
            allowed_input_types=[PropertyInputType(base_type="str", placeholder="secret-name")],
            hidden=False,
            required=True,
            use_in_key=False,
        ),
        ListItemPropertyAttribute(
            attribute_id="key",
            display_name="Secret Key",
            allowed_input_types=[PropertyInputType(base_type="str", placeholder="secret-key")],
            hidden=False,
            required=True,
            use_in_key=False,
        ),
    ]

    def __init__(self, env_var, name, key, **kwargs):
        self.env_var = env_var
        self.name = name
        self.key = key

    def get_all_validation_errors(self) -> List[str]:
        """Perform custom validation on an instance."""
        validation_errors = []
        if not self.env_var:
            validation_errors.append("Required environment variable was not specified.")
        if not self.name:
            validation_errors.append("Required secret name was not specified.")
        elif not is_valid_kubernetes_resource_name(self.name):
            validation_errors.append(
                f"Secret name '{self.name}' is not a valid Kubernetes resource name.",
            )
        if not self.key:
            validation_errors.append("Required secret key was not specified.")
        elif not is_valid_kubernetes_key(self.key):
            validation_errors.append(
                f"Key '{self.key}' is not a valid Kubernetes secret key.",
            )

        return validation_errors

    def add_to_execution_object(self, runtime_processor: RuntimePipelineProcessor, execution_object: Any, **kwargs):
        """Add KubernetesSecret instance to the execution object for the given runtime processor"""
        runtime_processor.add_kubernetes_secret(instance=self, execution_object=execution_object, **kwargs)


class VolumeMount(ElyraPropertyListItem):
    """An ElyraProperty representing a single PVC"""

    applies_to_generic = True
    applies_to_custom = True

    property_id = MOUNTED_VOLUMES
    property_display_name = "Data Volumes"
    property_description = """Volumes to be mounted in this node. The specified Persistent Volume Claims
    must exist in the Kubernetes namespace where the node is executed or this node will not run."""
    property_attributes = [
        ListItemPropertyAttribute(
            attribute_id="path",
            display_name="Mount Path",
            allowed_input_types=[PropertyInputType(base_type="str", placeholder="/mount/path")],
            hidden=False,
            required=True,
            use_in_key=True,
        ),
        ListItemPropertyAttribute(
            attribute_id="pvc_name",
            display_name="Persistent Volume Claim Name",
            allowed_input_types=[PropertyInputType(base_type="str", placeholder="pvc-name")],
            hidden=False,
            required=True,
            use_in_key=False,
        ),
        ListItemPropertyAttribute(
            attribute_id="sub_path",
            display_name="Sub Path",
            allowed_input_types=[PropertyInputType(base_type="str", placeholder="relative/path/within/volume")],
            hidden=False,
            required=False,
            use_in_key=False,
        ),
        ListItemPropertyAttribute(
            attribute_id="read_only",
            display_name="Mount volume read-only",
            allowed_input_types=[PropertyInputType(base_type="bool", default_value=False, placeholder=None)],
            hidden=False,
            required=False,
            use_in_key=False,
        ),
    ]

    def __init__(self, path: str, pvc_name: str, sub_path: str, read_only: bool, **kwargs):
        self.path = path
        self.pvc_name = pvc_name
        self.sub_path = sub_path
        self.read_only = read_only

    def get_all_validation_errors(self) -> List[str]:
        """Identify configuration issues for this instance"""
        validation_errors = []
        if not self.path:
            validation_errors.append("Required mount path was not specified.")
        if not self.pvc_name:
            validation_errors.append("Required persistent volume claim name was not specified.")
        elif not is_valid_kubernetes_resource_name(self.pvc_name):
            validation_errors.append(f"PVC name '{self.pvc_name}' is not a valid Kubernetes resource name.")
        if self.sub_path and self.sub_path.startswith("/"):
            validation_errors.append(f"Sub-path '{self.sub_path}' must be a relative path.")

        return validation_errors

    def add_to_execution_object(self, runtime_processor: RuntimePipelineProcessor, execution_object: Any, **kwargs):
        """Add VolumeMount instance to the execution object for the given runtime processor"""
        self.path = f"/{self.path.strip('/')}"  # normalize path
        if self.read_only is None:
            self.read_only = False
        runtime_processor.add_mounted_volume(instance=self, execution_object=execution_object, **kwargs)

    def should_discard(self) -> bool:
        """
        Returns a boolean indicating whether this VolumeMount instance should be silently discarded on
        the basis of its mount path, PVC name, and sub-path attribute values. If these attributes
        don't contain values this instance will not be validated or processed.
        """
        # ignore the read_only attribute because it always contains a value
        return not (self.path or self.pvc_name or self.sub_path)


class KubernetesAnnotation(ElyraPropertyListItem):
    """An ElyraProperty representing a single Kubernetes annotation"""

    applies_to_generic = True
    applies_to_custom = True

    property_id = KUBERNETES_POD_ANNOTATIONS
    property_display_name = "Kubernetes Pod Annotations"
    property_description = """Metadata to be added to this node. The metadata is exposed
    as annotation in the Kubernetes pod that executes this node."""
    property_attributes = [
        ListItemPropertyAttribute(
            attribute_id="key",
            display_name="Key",
            allowed_input_types=[PropertyInputType(base_type="str", placeholder="annotation_key")],
            hidden=False,
            required=True,
            use_in_key=True,
        ),
        ListItemPropertyAttribute(
            attribute_id="value",
            display_name="Value",
            allowed_input_types=[PropertyInputType(base_type="str", placeholder="annotation_value")],
            hidden=False,
            required=False,
            use_in_key=False,
        ),
    ]

    def __init__(self, key, value, **kwargs):
        self.key = key
        self.value = value

    def get_value_for_dict_entry(self) -> Union[str, Dict[str, Any]]:
        """
        Returns the value to be used when constructing a dict from a list of ElyraPropertyListItem.
        A KubernetesAnnotation dict entry will be of the form {self.key: self.value}
        """
        return self.value

    def get_all_validation_errors(self) -> List[str]:
        """Perform custom validation on an instance."""
        validation_errors = []
        # verify annotation key
        if not self.key:
            validation_errors.append("Required annotation key was not specified.")
        elif not is_valid_annotation_key(self.key):
            validation_errors.append(f"'{self.key}' is not a valid Kubernetes annotation key.")
        # verify annotation value
        if not is_valid_annotation_value(self.value):
            validation_errors.append(f"'{self.value}' is not a valid Kubernetes annotation value.")

        return validation_errors

    def add_to_execution_object(self, runtime_processor: RuntimePipelineProcessor, execution_object: Any, **kwargs):
        """Add KubernetesAnnotation instance to the execution object for the given runtime processor"""
        runtime_processor.add_kubernetes_pod_annotation(instance=self, execution_object=execution_object, **kwargs)


class KubernetesLabel(ElyraPropertyListItem):
    """An ElyraProperty representing a single Kubernetes pod label"""

    applies_to_generic = True
    applies_to_custom = True

    property_id = KUBERNETES_POD_LABELS
    property_display_name = "Kubernetes Pod Labels"
    property_description = """Metadata to be added to this node. The metadata is
    exposed as label in the Kubernetes pod that executes this node."""
    property_attributes = [
        ListItemPropertyAttribute(
            attribute_id="key",
            display_name="Key",
            allowed_input_types=[PropertyInputType(base_type="str", placeholder="label_key")],
            hidden=False,
            required=True,
            use_in_key=True,
        ),
        ListItemPropertyAttribute(
            attribute_id="value",
            display_name="Value",
            allowed_input_types=[PropertyInputType(base_type="str", placeholder="label_value")],
            hidden=False,
            required=False,
            use_in_key=False,
        ),
    ]

    def __init__(self, key, value, **kwargs):
        self.key = key
        self.value = value

    def get_value_for_dict_entry(self) -> Union[str, Dict[str, Any]]:
        """
        Returns the value to be used when constructing a dict from a list of ElyraPropertyListItem.
        A KubernetesLabel dict entry will be of the form {self.key: self.value}
        """
        return self.value

    def get_all_validation_errors(self) -> List[str]:
        """Perform custom validation on an instance."""
        validation_errors = []
        # verify label key
        if not self.key:
            validation_errors.append("Required label key was not specified.")
        elif not is_valid_label_key(self.key):
            validation_errors.append(f"'{self.key}' is not a valid Kubernetes label key.")
        # verify label value
        if not is_valid_label_value(self.value):
            validation_errors.append(f"'{self.value}' is not a valid Kubernetes label value.")
        return validation_errors

    def add_to_execution_object(self, runtime_processor: RuntimePipelineProcessor, execution_object: Any, **kwargs):
        """Add KubernetesLabel instance to the execution object for the given runtime processor"""
        runtime_processor.add_kubernetes_pod_label(instance=self, execution_object=execution_object, **kwargs)


class KubernetesToleration(ElyraPropertyListItem):
    """An ElyraProperty representing a single Kubernetes toleration"""

    applies_to_generic = True
    applies_to_custom = True

    property_id = KUBERNETES_TOLERATIONS
    property_display_name = "Kubernetes Tolerations"
    property_description = "Kubernetes tolerations to apply to the pod where the node is executed."
    property_attributes = [
        ListItemPropertyAttribute(
            attribute_id="key",
            display_name="Key",
            allowed_input_types=[PropertyInputType(base_type="str", placeholder="key")],
            hidden=False,
            required=False,
            use_in_key=True,
        ),
        ListItemPropertyAttribute(
            attribute_id="operator",
            display_name="Operator",
            allowed_input_types=[PropertyInputType(base_type="str", default_value="Equal", enum=["Equal", "Exists"])],
            hidden=False,
            required=True,
            use_in_key=True,
        ),
        ListItemPropertyAttribute(
            attribute_id="value",
            display_name="Value",
            allowed_input_types=[PropertyInputType(base_type="str", placeholder="value")],
            hidden=False,
            required=False,
            use_in_key=True,
        ),
        ListItemPropertyAttribute(
            attribute_id="effect",
            display_name="Effect",
            allowed_input_types=[
                PropertyInputType(
                    base_type="str",
                    placeholder="NoSchedule",
                    enum=["", "NoExecute", "NoSchedule", "PreferNoSchedule"],
                )
            ],
            hidden=False,
            required=False,
            use_in_key=True,
        ),
    ]

    def __init__(self, key, operator, value, effect, **kwargs):
        self.key = key
        self.operator = operator
        self.value = value
        self.effect = effect

    def get_all_validation_errors(self) -> List[str]:
        """
        Perform custom validation on an instance using the constraints documented in
        https://kubernetes.io/docs/concepts/scheduling-eviction/taint-and-toleration/
        """
        validation_errors = []

        # Ensure the operator is valid
        if self.operator not in ["Exists", "Equal"]:
            validation_errors.append(
                f"'{self.operator}' is not a valid operator: the value must be one of 'Exists' or 'Equal'."
            )

        if self.operator == "Equal" and not self.key:
            validation_errors.append(
                f"'{self.operator}' is not a valid operator: operator must be 'Exists' if no key is specified."
            )

        if (
            self.effect is not None
            and len(self.effect) > 0
            and self.effect not in ["NoExecute", "NoSchedule", "PreferNoSchedule"]
        ):
            validation_errors.append(
                f"'{self.effect}' is not a valid effect: effect must be one "
                f"of 'NoExecute', 'NoSchedule', or 'PreferNoSchedule'."
            )

        if self.operator == "Exists" and self.value:
            validation_errors.append(
                f"'{self.value}' is not a valid value: value should be empty if operator is 'Exists'."
            )
        return validation_errors

    def add_to_execution_object(self, runtime_processor: RuntimePipelineProcessor, execution_object: Any, **kwargs):
        """Add KubernetesToleration instance to the execution object for the given runtime processor"""
        runtime_processor.add_kubernetes_toleration(instance=self, execution_object=execution_object, **kwargs)


class ElyraPropertyList(list):
    """
    A list class that exposes functionality specific to lists whose entries are
    of the class ElyraOwnedPropertyListItem.
    """

    def to_dict(self: List[ElyraPropertyListItem], use_prop_as_value: bool = False) -> Dict[str, str]:
        """
        Each Elyra-owned property consists of a set of attributes, some subset of which represents
        a unique key. Lists of these properties, however, often need converted to dictionary
        form for processing - so we must convert.
        """
        prop_dict = {}
        for prop in self:
            if prop is None or not isinstance(prop, ElyraPropertyListItem):
                continue  # invalid entry; skip inclusion and continue
            prop_key = prop.get_key_for_dict_entry()
            if prop_key is None:
                continue  # invalid entry; skip inclusion and continue

            prop_value = prop.get_value_for_dict_entry()
            if use_prop_as_value:
                prop_value = prop  # use the property object itself as the value
            prop_dict[prop_key] = prop_value

        return prop_dict

    def deduplicate(self: ElyraPropertyList) -> ElyraPropertyList:
        """Remove duplicates from the given list"""
        instance_dict = self.to_dict(use_prop_as_value=True)
        return ElyraPropertyList({**instance_dict}.values())

    @staticmethod
    def merge(primary: ElyraPropertyList, secondary: ElyraPropertyList) -> ElyraPropertyList:
        """
        Merge two lists of Elyra-owned properties, preferring the values given in the
        primary parameter in the case of a matching key between the two lists.
        """
        primary_dict = primary.to_dict(use_prop_as_value=True)
        secondary_dict = secondary.to_dict(use_prop_as_value=True)

        merged_list = list({**secondary_dict, **primary_dict}.values())
        return ElyraPropertyList(merged_list)

    @staticmethod
    def difference(minuend: ElyraPropertyList, subtrahend: ElyraPropertyList) -> ElyraPropertyList:
        """
        Given two lists of Elyra-owned properties, remove any duplicate instances
        found in the second (subtrahend) from the first (minuend), if present.

        :param minuend: list to be subtracted from
        :param subtrahend: list from which duplicates will be determined and given preference

        :returns: the difference of the two lists
        """
        subtract_dict = minuend.to_dict(use_prop_as_value=True)
        for key in subtrahend.to_dict().keys():
            if key in subtract_dict:
                subtract_dict.pop(key)

        return ElyraPropertyList(subtract_dict.values())

    def add_to_execution_object(self, runtime_processor: RuntimePipelineProcessor, execution_object: Any):
        """
        Add a property instance to the execution object for the given runtime processor
        for each list item.
        """
        for item in self:
            if isinstance(item, ElyraPropertyListItem):
                item.add_to_execution_object(runtime_processor=runtime_processor, execution_object=execution_object)


class ElyraPropertyJSONEncoder(json.JSONEncoder):
    """
    A JSON Encoder class to prevent errors during serialization of Elyra-owned property classes.
    """

    def default(self, o):
        """
        Render dataclass content as dict
        """
        return o.__dict__ if isinstance(o, (ElyraProperty, PropertyInputType)) else super().default(o)


class ComponentProperty(object):
    """
    Represents a single property for a pipeline component
    """

    def __init__(
        self,
        id: str,
        name: str,
        json_data_type: str,
        description: str,
        allowed_input_types: List[Optional[str]],
        value: Optional[Any] = "",
        required: Optional[bool] = False,
        allow_no_options: Optional[bool] = False,
        parsed_data_type: Optional[str] = None,
        items: Optional[List[str]] = None,
    ):
        """
        :param id: Unique identifier for a property
        :param name: The name of the property for display
        :param json_data_type: The JSON data type that represents this property's value
        :param allowed_input_types: The input types that the property can accept, including those for custom rendering
        :param value: The default value of the property
        :param description: A description of the property for display
        :param required: Whether the property is required
        :param allow_no_options: Specifies whether to allow parent nodes that don't specifically
            define output properties to be selected as input to this node property
        :param parsed_data_type: The raw data typed exactly as parsed from the component spec
        :param items: For properties with a control of 'EnumControl', the items making up the enum
        """

        if not id:
            raise ValueError("Invalid component: Missing field 'id'.")
        if not name:
            raise ValueError("Invalid component: Missing field 'name'.")

        self._ref = id
        self._name = name
        self._json_data_type = json_data_type

        # The JSON type that the value entered for this property will be rendered in.
        # E.g., array types are entered by users and processed by the backend as
        # strings whereas boolean types are entered and processed as booleans
        self._value_entry_type = json_data_type
        if json_data_type in {"array", "object"}:
            self._value_entry_type = "string"

        if json_data_type == "boolean" and isinstance(value, str):
            value = value in ["True", "true"]
        elif json_data_type == "number" and isinstance(value, str):
            try:
                # Attempt to coerce string to integer value
                value = int(value)
            except ValueError:
                # Value could not be coerced to integer, assume float
                value = float(value)
        if json_data_type in {"array", "object"} and not isinstance(value, str):
            value = str(value)
        self._value = value

        self._description = description
        self._parsed_data_type = parsed_data_type
        self._allowed_input_types = allowed_input_types
        self._items = items or []

        # Check description for information about 'required' property
        if "not optional" in description.lower() or (
            "required" in description.lower()
            and "not required" not in description.lower()
            and "n't required" not in description.lower()
        ):
            required = True

        self._required = required
        self._allow_no_options = allow_no_options

    @property
    def ref(self) -> str:
        return self._ref

    @property
    def name(self) -> str:
        return self._name

    @property
    def allowed_input_types(self) -> List[Optional[str]]:
        return self._allowed_input_types

    @property
    def json_data_type(self) -> str:
        return self._json_data_type

    @property
    def value_entry_type(self) -> str:
        return self._value_entry_type

    @property
    def value(self) -> Any:
        return self._value

    @property
    def description(self) -> str:
        return self._description

    @property
    def items(self) -> List[str]:
        return self._items

    @property
    def required(self) -> bool:
        return bool(self._required)

    @property
    def allow_no_options(self) -> bool:
        return self._allow_no_options

    @property
    def parsed_data_type(self) -> Optional[str]:
        return self._parsed_data_type

    @staticmethod
    def render_property_details(prop: ComponentProperty) -> str:
        """
        Render the property data type and UI hints needed for the specified property for
        use in the custom component properties DAG template
        :returns: a string literal containing the JSON object to be rendered in the DAG
        """
        json_dict = {"title": prop.name, "description": prop.description}
        if len(prop.allowed_input_types) == 1:
            # Property only accepts a single type of input
            input_type = prop.allowed_input_types[0]
            if not input_type:
                # This is an output
                json_dict["type"] = "string"
                json_dict["uihints"] = {"ui:widget": "hidden", "outputpath": True}
            elif input_type == "inputpath":
                json_dict.update(
                    {
                        "type": "object",
                        "properties": {
                            "widget": {"type": "string", "default": input_type},
                            "value": {"type": "string", "enum": []},
                        },
                        "uihints": {"widget": {"ui:field": "hidden"}, "value": {input_type: True}},
                    }
                )
            elif input_type == "file":
                json_dict["type"] = "string"
                json_dict["uihints"] = {"ui:widget": input_type}
            else:
                json_dict["type"] = prop.value_entry_type

                # Render default value if it is not None
                if prop.value is not None:
                    json_dict["default"] = prop.value
        else:
            # Property accepts multiple types of inputs; render a oneOf block
            one_of = []
            for widget_type in prop.allowed_input_types:
                obj = {
                    "type": "object",
                    "properties": {"widget": {"type": "string"}, "value": {}},
                    "uihints": {"widget": {"ui:widget": "hidden"}, "value": {}},
                }
                if widget_type == "inputvalue":
                    obj["title"] = InputTypeDescriptionMap[prop.value_entry_type].value
                    obj["properties"]["widget"]["default"] = prop.value_entry_type
                    obj["properties"]["value"]["type"] = prop.value_entry_type
                    if prop.value_entry_type == "boolean":
                        obj["properties"]["value"]["title"] = " "

                    # Render default value if it is not None
                    if prop.value is not None:
                        obj["properties"]["value"]["default"] = prop.value
                else:  # custom widget types (inputpath, file, parameter)
                    obj["title"] = InputTypeDescriptionMap[widget_type].value
                    obj["properties"]["widget"]["default"] = widget_type
                    obj["properties"]["value"]["type"] = "string"
                    if widget_type == "outputpath":
                        obj["uihints"]["value"] = {"ui:readonly": "true", widget_type: True}
                        obj["properties"]["value"]["type"] = "string"
                    elif widget_type == "inputpath":
                        obj["uihints"]["value"] = {widget_type: True}
                        obj["properties"]["value"]["enum"] = []
                        if prop.allow_no_options:
                            obj["uihints"]["allownooptions"] = prop.allow_no_options
                    elif widget_type == "parameter":
                        # Render parsed data type if present
                        if prop.parsed_data_type:
                            obj["uihints"]["value"] = {"ui:typefilter": prop.parsed_data_type}
                    else:
                        obj["uihints"]["value"] = {"ui:widget": widget_type}

                one_of.append(obj)
            json_dict["oneOf"] = one_of

        return json.dumps(json_dict)


class InputTypeDescriptionMap(Enum):
    """A mapping of input types to the description that will appear in the UI"""

    string = "Please enter a string value:"
    number = "Please enter a number value:"
    boolean = "Please select or deselect the checkbox:"
    file = "Please select a file to use as input:"
    inputpath = "Please select an output from a parent:"
    outputpath = None  # outputs are read-only and don't require a description
    parameter = "Please select a parameter to use as input:"


class PipelineParameter(ElyraPropertyListItem):
    """An ElyraProperty representing a single pipeline parameter"""

    property_id = PIPELINE_PARAMETERS
    property_display_name = "Pipeline Parameters"
    property_description = """Typed variables to be passed to the pipeline."""

    def __init__(self, name, description, value, default_value, required, **kwargs):
        self.name = name
        self.description = description
        self.required = required or False

        # Assign selected type, if given
        self.selected_type = default_value.get("type") if isinstance(default_value, dict) else None

        # Assign default value, if given
        self.default_value = None
        if isinstance(default_value, dict):
            if default_value.get("value") not in [None, ""]:
                self.default_value = default_value.get("value")

        # Assign value, if set; use default_value if not
        self.value = value if value not in [None, ""] else self.default_value

    def get_value_for_dict_entry(self) -> Union[str, Dict[str, Any]]:
        """
        Returns the value to be used when constructing a dict from a list of ElyraPropertyListItem.
        A PipelineParameter dict entry will be of the form: {self.name: self.value}
        """
        return self.value

    def to_dict(self) -> Dict[str, Any]:
        """Convert instance to a dict with relevant class attributes."""
        return {
            "name": self.name,
            "description": self.description,
            "value": self.value,
            "default_value": self.default_value,
            "type": self.selected_type,
            "required": self.required,
        }

    def get_value_for_display(self) -> Dict[str, Any]:
        """Get a representation of the instance to display in UI error messages."""
        return self.to_dict()

    def get_all_validation_errors(self) -> List[str]:
        """Perform custom validation on an instance."""
        raise RuntimeError("Method should be invoked only by runtime-specific subclasses.")

    def add_to_execution_object(self, runtime_processor: RuntimePipelineProcessor, execution_object: Any, **kwargs):
        raise RuntimeError("Method should not be invoked.")



#
# Copyright 2018-2023 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import hashlib
import json
import logging
import os
from pathlib import Path
import subprocess
from subprocess import CalledProcessError
from subprocess import CompletedProcess
from subprocess import run
import sys
from tempfile import TemporaryFile
import time
from typing import Optional

import minio
import mock
import nbformat
import papermill
import pytest

from elyra.kfp import bootstrapper

# To run this test from an IDE:
# 1. set PYTHONPATH='`path-to-repo`/etc/docker-scripts' and working directory to `path-to-repo`
# 2. Manually launch test_minio container: docker run --name test_minio -d -p 9000:9000 minio/minio server /data
#    (this is located in Makefile)
#
# NOTE: Any changes to elyra/tests/kfp/resources/test-notebookA.ipynb require an
# update of elyra/tests/kfp/resources/test-archive.tgz  using the command below:
# tar -cvzf test-archive.tgz test-notebookA.ipynb


MINIO_HOST_PORT = os.getenv("MINIO_HOST_PORT", "127.0.0.1:9000")

ELYRA_ROOT_DIR = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))
RESOURCES_DIR = os.path.join(ELYRA_ROOT_DIR, "elyra", "tests", "kfp", "resources")


@pytest.fixture(scope="module", autouse=True)
def start_minio():
    """Start the minio container to simulate COS."""

    # The docker run command will fail if an instance of the test_minio container is running.
    # We'll make a "silent" attempt to start.  If that fails, assume its due to the container
    # conflict, force its shutdown, and try once more.  If successful, yield the minio instance
    # but also shutdown on the flip-side of the yield (when the fixture is cleaned up).
    #
    # Although actions like SIGINT (ctrl-C) should still trigger the post-yield logic, urgent
    # interrupts like SIGQUIT (ctrl-\) or multiple SIGINTs can still orphan the container, so
    # we still need the pre-yield behavior.

    minio = start_minio_container(False)
    if minio is None:  # Got a failure. Shutdown (assumed) container and try once more.
        stop_minio_container()
        minio = start_minio_container(True)

    time.sleep(3)  # give container a chance to start
    yield minio
    stop_minio_container()


def start_minio_container(raise_on_failure: bool = False) -> Optional[CompletedProcess]:
    minio = None
    try:
        minio = run(
            ["docker", "run", "--name", "test_minio", "-d", "-p", "9000:9000", "minio/minio", "server", "/data"],
            check=True,
        )
    except CalledProcessError as ex:
        if raise_on_failure:
            raise RuntimeError(f"Error executing docker process: {ex}") from ex

    return minio


def stop_minio_container():
    run(["docker", "rm", "-f", "test_minio"], check=True)


@pytest.fixture(scope="function")
def s3_setup():
    bucket_name = "test-bucket"
    cos_client = minio.Minio(MINIO_HOST_PORT, access_key="minioadmin", secret_key="minioadmin", secure=False)
    cos_client.make_bucket(bucket_name)

    yield cos_client

    cleanup_files = cos_client.list_objects(bucket_name, recursive=True)
    for file in cleanup_files:
        cos_client.remove_object(bucket_name, file.object_name)
    cos_client.remove_bucket(bucket_name)


def main_method_setup_execution(monkeypatch, s3_setup, tmpdir, argument_dict):
    """Primary body for main method testing..."""
    monkeypatch.setattr(bootstrapper.OpUtil, "parse_arguments", lambda x: argument_dict)
    monkeypatch.setattr(bootstrapper.OpUtil, "package_install", mock.Mock(return_value=True))

    monkeypatch.setenv("AWS_ACCESS_KEY_ID", "minioadmin")
    monkeypatch.setenv("AWS_SECRET_ACCESS_KEY", "minioadmin")
    monkeypatch.setenv("TEST_ENV_VAR1", "test_env_var1")

    s3_setup.fput_object(
        bucket_name=argument_dict["cos-bucket"],
        object_name="test-directory/test-file.txt",
        file_path=os.path.join(RESOURCES_DIR, "test-requirements-elyra.txt"),
    )
    s3_setup.fput_object(
        bucket_name=argument_dict["cos-bucket"],
        object_name="test-directory/test,file.txt",
        file_path=os.path.join(RESOURCES_DIR, "test-bad-requirements-elyra.txt"),
    )
    s3_setup.fput_object(
        bucket_name=argument_dict["cos-bucket"],
        object_name="test-directory/test-archive.tgz",
        file_path=os.path.join(RESOURCES_DIR, "test-archive.tgz"),
    )

    with tmpdir.as_cwd():
        bootstrapper.main()
        test_file_list = [
            "test-archive.tgz",
            "test-file.txt",
            "test,file.txt",
            "test-file/test-file-copy.txt",
            "test-file/test,file/test,file-copy.txt",
            "test-notebookA.ipynb",
            "test-notebookA-output.ipynb",
            "test-notebookA.html",
        ]
        # Ensure working directory has all the files.
        for file in test_file_list:
            assert os.path.isfile(file)
        # Ensure upload directory has all the files EXCEPT the output notebook
        # since it was it is uploaded as the input notebook (test-notebookA.ipynb)
        # (which is included in the archive at start).
        for file in test_file_list:
            if file != "test-notebookA-output.ipynb":
                assert s3_setup.stat_object(
                    bucket_name=argument_dict["cos-bucket"], object_name="test-directory/" + file
                )
                if file == "test-notebookA.html":
                    with open("test-notebookA.html") as html_file:
                        assert "TEST_ENV_VAR1: test_env_var1" in html_file.read()


def _get_operation_instance(monkeypatch, s3_setup):
    config = {
        "cos-endpoint": "http://" + MINIO_HOST_PORT,
        "cos-user": "minioadmin",
        "cos-password": "minioadmin",
        "cos-bucket": "test-bucket",
        "filepath": "untitled.ipynb",
    }

    op = bootstrapper.FileOpBase.get_instance(**config)

    # use the same minio instance used by the test
    # to avoid access denied errors when two minio
    # instances exist
    monkeypatch.setattr(op, "cos_client", s3_setup)

    return op


def test_main_method(monkeypatch, s3_setup, tmpdir):
    argument_dict = {
        "cos-endpoint": "http://" + MINIO_HOST_PORT,
        "cos-bucket": "test-bucket",
        "cos-directory": "test-directory",
        "cos-dependencies-archive": "test-archive.tgz",
        "filepath": os.path.join(RESOURCES_DIR, "test-notebookA.ipynb"),
        "inputs": "test-file.txt;test,file.txt",
        "outputs": "test-file/test-file-copy.txt;test-file/test,file/test,file-copy.txt",
        "user-volume-path": None,
    }
    main_method_setup_execution(monkeypatch, s3_setup, tmpdir, argument_dict)


def test_main_method_with_wildcard_outputs(monkeypatch, s3_setup, tmpdir):
    argument_dict = {
        "cos-endpoint": "http://" + MINIO_HOST_PORT,
        "cos-bucket": "test-bucket",
        "cos-directory": "test-directory",
        "cos-dependencies-archive": "test-archive.tgz",
        "filepath": os.path.join(RESOURCES_DIR, "test-notebookA.ipynb"),
        "inputs": "test-file.txt;test,file.txt",
        "outputs": "test-file/*",
        "user-volume-path": None,
    }
    main_method_setup_execution(monkeypatch, s3_setup, tmpdir, argument_dict)


def test_main_method_with_dir_outputs(monkeypatch, s3_setup, tmpdir):
    argument_dict = {
        "cos-endpoint": "http://" + MINIO_HOST_PORT,
        "cos-bucket": "test-bucket",
        "cos-directory": "test-directory",
        "cos-dependencies-archive": "test-archive.tgz",
        "filepath": os.path.join(RESOURCES_DIR, "test-notebookA.ipynb"),
        "inputs": "test-file.txt;test,file.txt",
        "outputs": "test-file",  # this is the directory that contains the outputs
        "user-volume-path": None,
    }
    main_method_setup_execution(monkeypatch, s3_setup, tmpdir, argument_dict)


def is_writable_dir(path):
    """Helper method determines whether 'path' is a writable directory"""
    try:
        with TemporaryFile(mode="w", dir=path) as t:
            t.write("1")
        return True
    except Exception:
        return False


def remove_file(filename, fail_ok=True):
    """Removes filename. If fail_ok is False an assert is raised
    if removal failed for any reason, e.g. filenotfound
    """
    try:
        os.remove(filename)
    except OSError as ose:
        if fail_ok is False:
            raise AssertionError(f"Cannot remove {filename}: {str(ose)} {ose}")


def test_process_metrics_method_not_writable_dir(monkeypatch, s3_setup, tmpdir):
    """Test for process_metrics_and_metadata

    Validates that the method can handle output directory that is not writable
    """

    # remove "default" output file if it already exists
    output_metadata_file = Path("/tmp") / "mlpipeline-ui-metadata.json"
    remove_file(output_metadata_file)

    try:
        monkeypatch.setenv("ELYRA_WRITABLE_CONTAINER_DIR", "/good/time/to/fail")
        argument_dict = {
            "cos-endpoint": f"http://{MINIO_HOST_PORT}",
            "cos-bucket": "test-bucket",
            "cos-directory": "test-directory",
            "cos-dependencies-archive": "test-archive.tgz",
            "filepath": os.path.join(RESOURCES_DIR, "test-notebookA.ipynb"),
            "inputs": "test-file.txt;test,file.txt",
            "outputs": "test-file/test-file-copy.txt;test-file/test,file/test,file-copy.txt",
            "user-volume-path": None,
        }
        main_method_setup_execution(monkeypatch, s3_setup, tmpdir, argument_dict)
    except Exception as ex:
        print(f"Writable dir test failed: {str(ex)} {ex}")
        assert False
    assert output_metadata_file.exists() is False


def test_process_metrics_method_no_metadata_file(monkeypatch, s3_setup, tmpdir):
    """Test for process_metrics_and_metadata

    Verifies that the method produces a valid KFP UI metadata file if
    the node's script | notebook did not generate this metadata file.
    """
    argument_dict = {
        "cos-endpoint": "http://" + MINIO_HOST_PORT,
        "cos-bucket": "test-bucket",
        "cos-directory": "test-directory",
        "cos-dependencies-archive": "test-archive.tgz",
        "filepath": os.path.join(RESOURCES_DIR, "test-notebookA.ipynb"),
        "inputs": "test-file.txt;test,file.txt",
        "outputs": "test-file/test-file-copy.txt;test-file/test,file/test,file-copy.txt",
        "user-volume-path": None,
    }

    output_path = Path(tmpdir)
    # metadata file name and location
    metadata_file = output_path / "mlpipeline-ui-metadata.json"
    # remove file if it already exists
    remove_file(metadata_file)

    # override the default output directory to make this test platform
    # independent
    monkeypatch.setenv("ELYRA_WRITABLE_CONTAINER_DIR", str(tmpdir))
    main_method_setup_execution(monkeypatch, s3_setup, tmpdir, argument_dict)

    # process_metrics should have generated a file named mlpipeline-ui-metadata.json
    # in tmpdir

    try:
        with open(metadata_file, "r") as f:
            metadata = json.load(f)
            assert metadata.get("outputs") is not None
            assert isinstance(metadata["outputs"], list)
            assert len(metadata["outputs"]) == 1
            assert metadata["outputs"][0]["storage"] == "inline"
            assert metadata["outputs"][0]["type"] == "markdown"
            assert (
                f"{argument_dict['cos-endpoint']}/{argument_dict['cos-bucket']}/{argument_dict['cos-directory']}"
                in metadata["outputs"][0]["source"]
            )
            assert argument_dict["cos-dependencies-archive"] in metadata["outputs"][0]["source"]
    except AssertionError:
        raise
    except Exception as ex:
        # Potential reasons for failures:
        # file not found, invalid JSON
        print(f'Validation of "{str(ex)}" failed: {ex}')
        assert False


def test_process_metrics_method_valid_metadata_file(monkeypatch, s3_setup, tmpdir):
    """Test for process_metrics_and_metadata

    Verifies that the method produces a valid KFP UI metadata file if
    the node's script | notebook generated this metadata file.
    """
    argument_dict = {
        "cos-endpoint": "http://" + MINIO_HOST_PORT,
        "cos-bucket": "test-bucket",
        "cos-directory": "test-directory",
        "cos-dependencies-archive": "test-archive.tgz",
        "filepath": os.path.join(RESOURCES_DIR, "test-notebookA.ipynb"),
        "inputs": "test-file.txt;test,file.txt",
        "outputs": "test-file/test-file-copy.txt;test-file/test,file/test,file-copy.txt",
        "user-volume-path": None,
    }

    output_path = Path(tmpdir)
    # metadata file name and location
    input_metadata_file = "mlpipeline-ui-metadata.json"
    output_metadata_file = output_path / input_metadata_file
    # remove output_metadata_file if it already exists
    remove_file(output_metadata_file)

    #
    # Simulate some custom metadata that the script | notebook produced
    #
    custom_metadata = {
        "some_property": "some property value",
        "outputs": [{"source": "gs://project/bucket/file.md", "type": "markdown"}],
    }

    with tmpdir.as_cwd():
        with open(input_metadata_file, "w") as f:
            json.dump(custom_metadata, f)
    # override the default output directory to make this test platform
    # independent
    monkeypatch.setenv("ELYRA_WRITABLE_CONTAINER_DIR", str(tmpdir))
    main_method_setup_execution(monkeypatch, s3_setup, tmpdir, argument_dict)

    # output_metadata_file should now exist

    try:
        with open(output_metadata_file, "r") as f:
            metadata = json.load(f)
            assert metadata.get("some_property") is not None
            assert metadata["some_property"] == custom_metadata["some_property"]
            assert metadata.get("outputs") is not None
            assert isinstance(metadata["outputs"], list)
            assert len(metadata["outputs"]) == 2
            for output in metadata["outputs"]:
                if output.get("storage") is not None:
                    assert output["storage"] == "inline"
                    assert output["type"] == "markdown"
                    assert (
                        f"{argument_dict['cos-endpoint']}/{argument_dict['cos-bucket']}/{argument_dict['cos-directory']}"  # noqa
                        in output["source"]
                    )
                    assert argument_dict["cos-dependencies-archive"] in output["source"]
                else:
                    assert output["type"] == custom_metadata["outputs"][0]["type"]
                    assert output["source"] == custom_metadata["outputs"][0]["source"]
    except AssertionError:
        raise
    except Exception as ex:
        # Potential reasons for failures:
        # file not found, invalid JSON
        print(f'Validation of "{str(ex)}" failed: {ex}')
        assert False


def test_process_metrics_method_invalid_metadata_file(monkeypatch, s3_setup, tmpdir):
    """Test for process_metrics_and_metadata

    Verifies that the method produces a valid KFP UI metadata file if
    the node's script | notebook generated an invalid metadata file.
    """
    argument_dict = {
        "cos-endpoint": f"http://{MINIO_HOST_PORT}",
        "cos-bucket": "test-bucket",
        "cos-directory": "test-directory",
        "cos-dependencies-archive": "test-archive.tgz",
        "filepath": os.path.join(RESOURCES_DIR, "test-notebookA.ipynb"),
        "inputs": "test-file.txt;test,file.txt",
        "outputs": "test-file/test-file-copy.txt;test-file/test,file/test,file-copy.txt",
        "user-volume-path": None,
    }

    output_path = Path(tmpdir)
    # metadata file name and location
    input_metadata_file = "mlpipeline-ui-metadata.json"
    output_metadata_file = output_path / input_metadata_file
    # remove output_metadata_file if it already exists
    remove_file(output_metadata_file)

    #
    # Populate the metadata file with some custom data that's not JSON
    #

    with tmpdir.as_cwd():
        with open(input_metadata_file, "w") as f:
            f.write("I am not a valid JSON data structure")
            f.write("1,2,3,4,5,6,7")

    # override the default output directory to make this test platform
    # independent
    monkeypatch.setenv("ELYRA_WRITABLE_CONTAINER_DIR", str(tmpdir))
    main_method_setup_execution(monkeypatch, s3_setup, tmpdir, argument_dict)

    # process_metrics replaces the existing metadata file
    # because its content cannot be merged

    try:
        with open(output_metadata_file, "r") as f:
            metadata = json.load(f)
            assert metadata.get("outputs") is not None
            assert isinstance(metadata["outputs"], list)
            assert len(metadata["outputs"]) == 1
            assert metadata["outputs"][0]["storage"] == "inline"
            assert metadata["outputs"][0]["type"] == "markdown"
            assert (
                f"{argument_dict['cos-endpoint']}/{argument_dict['cos-bucket']}/{argument_dict['cos-directory']}"
                in metadata["outputs"][0]["source"]
            )
            assert argument_dict["cos-dependencies-archive"] in metadata["outputs"][0]["source"]
    except AssertionError:
        raise
    except Exception as ex:
        # Potential reasons for failures:
        # file not found, invalid JSON
        print(f'Validation of "{str(ex)}" failed: {ex}')
        assert False


def test_fail_bad_notebook_main_method(monkeypatch, s3_setup, tmpdir):
    argument_dict = {
        "cos-endpoint": f"http://{MINIO_HOST_PORT}",
        "cos-bucket": "test-bucket",
        "cos-directory": "test-directory",
        "cos-dependencies-archive": "test-bad-archiveB.tgz",
        "filepath": os.path.join(RESOURCES_DIR, "test-bad-notebookB.ipynb"),
        "inputs": "test-file.txt",
        "outputs": "test-file/test-copy-file.txt",
        "user-volume-path": None,
    }

    monkeypatch.setattr(bootstrapper.OpUtil, "parse_arguments", lambda x: argument_dict)
    monkeypatch.setattr(bootstrapper.OpUtil, "package_install", mock.Mock(return_value=True))

    mocked_func = mock.Mock(
        return_value="default",
        side_effect=[
            "test-bad-archiveB.tgz",
            "test-file.txt",
            "test-bad-notebookB-output.ipynb",
            "test-bad-notebookB.html",
            "test-file.txt",
        ],
    )
    monkeypatch.setattr(bootstrapper.FileOpBase, "get_object_storage_filename", mocked_func)

    monkeypatch.setenv("AWS_ACCESS_KEY_ID", "minioadmin")
    monkeypatch.setenv("AWS_SECRET_ACCESS_KEY", "minioadmin")

    s3_setup.fput_object(
        bucket_name=argument_dict["cos-bucket"],
        object_name="test-file.txt",
        file_path=os.path.join(ELYRA_ROOT_DIR, "README.md"),
    )
    s3_setup.fput_object(
        bucket_name=argument_dict["cos-bucket"],
        object_name="test-bad-archiveB.tgz",
        file_path=os.path.join(RESOURCES_DIR, "test-bad-archiveB.tgz"),
    )

    with tmpdir.as_cwd():
        with pytest.raises(papermill.exceptions.PapermillExecutionError):
            bootstrapper.main()


def test_package_installation(monkeypatch, virtualenv):
    elyra_packages = {
        "ipykernel": "5.3.0",
        "ansiwrap": "0.8.4",
        "packaging": "20.0",
    }
    current_packages = {
        "bleach": "3.1.5",
        "ansiwrap": "0.7.0",
        "packaging": "20.4",
    }
    expected_packages = {
        "ipykernel": "5.3.0",
        "ansiwrap": "0.8.4",
        "packaging": "20.4",
    }

    mocked_func = mock.Mock(side_effect=[elyra_packages, current_packages])

    monkeypatch.setattr(bootstrapper.OpUtil, "package_list_to_dict", mocked_func)
    monkeypatch.setattr(sys, "executable", virtualenv.python)

    virtualenv.run("python3 -m pip install bleach==3.1.5")
    virtualenv.run("python3 -m pip install ansiwrap==0.7.0")
    virtualenv.run("python3 -m pip install packaging==20.4")

    bootstrapper.OpUtil.package_install(user_volume_path=None)
    virtualenv_packages = {}
    output = virtualenv.run("python3 -m pip freeze", capture=True)
    print("This is the [pip freeze] output :\n" + output)
    for line in output.strip().split("\n"):
        if " @ " in line:
            package_name, package_version = line.strip("\n").split(sep=" @ ")
        elif "===" in line:
            package_name, package_version = line.strip("\n").split(sep="===")
        else:
            package_name, package_version = line.strip("\n").split(sep="==")
        virtualenv_packages[package_name] = package_version

    for package, version in expected_packages.items():
        assert virtualenv_packages[package] == version


@pytest.mark.skip(reason="Temporarily disabling see elyra#3072")
def test_package_installation_with_target_path(monkeypatch, virtualenv, tmpdir):
    # TODO : Need to add test for direct-source e.g. ' @ '
    elyra_packages = {
        "ipykernel": "5.3.0",
        "ansiwrap": "0.8.4",
        "packaging": "22.0",
    }
    current_packages = {
        "bleach": "3.1.5",
        "ansiwrap": "0.7.0",
        "packaging": "21.0",
    }
    expected_packages = {
        "ipykernel": "5.3.0",
        "ansiwrap": "0.8.4",
        "packaging": "22.0",
    }

    mocked_func = mock.Mock(side_effect=[elyra_packages, current_packages])

    monkeypatch.setattr(bootstrapper.OpUtil, "package_list_to_dict", mocked_func)
    monkeypatch.setattr(sys, "executable", virtualenv.python)

    virtualenv.run("python3 -m pip install --upgrade pip")
    for package, version in current_packages.items():
        virtualenv.run(f"python3 -m pip install --target={tmpdir} {package}=={version}")

    bootstrapper.OpUtil.package_install(user_volume_path=str(tmpdir))
    virtualenv_packages = {}
    output = virtualenv.run(f"python3 -m pip freeze --path={tmpdir}", capture=True)
    print("This is the [pip freeze] output :\n" + output)
    for line in output.strip().split("\n"):
        if " @ " in line:
            package_name, package_version = line.strip("\n").split(sep=" @ ")
        elif "===" in line:
            package_name, package_version = line.strip("\n").split(sep="===")
        else:
            package_name, package_version = line.strip("\n").split(sep="==")
        virtualenv_packages[package_name] = package_version

    for package, version in expected_packages.items():
        assert (
            virtualenv_packages[package].split(".")[0] == version.split(".")[0]
        ), f"Major version mismatch for package {package}"


def test_convert_notebook_to_html(tmpdir):
    notebook_file = os.path.join(RESOURCES_DIR, "test-notebookA.ipynb")
    notebook_output_html_file = "test-notebookA.html"

    with tmpdir.as_cwd():
        bootstrapper.NotebookFileOp.convert_notebook_to_html(notebook_file, notebook_output_html_file)

        assert os.path.isfile(notebook_output_html_file)
        # Validate that an html file got generated from the notebook
        with open(notebook_output_html_file, "r") as html_file:
            html_data = html_file.read()
            assert html_data.startswith("<!DOCTYPE html>")
            assert "TEST_ENV_VAR1" in html_data  # from os.getenv("TEST_ENV_VAR1")
            assert html_data.endswith("</html>\n")


def test_fail_convert_notebook_to_html(tmpdir):
    notebook_file = os.path.join(RESOURCES_DIR, "test-bad-notebookA.ipynb")
    notebook_output_html_file = "bad-notebookA.html"
    with tmpdir.as_cwd():
        # Recent versions raising typeError due to #1130
        # https://github.com/jupyter/nbconvert/pull/1130
        with pytest.raises((TypeError, nbformat.validator.NotebookValidationError)):
            bootstrapper.NotebookFileOp.convert_notebook_to_html(notebook_file, notebook_output_html_file)


def test_get_file_object_store(monkeypatch, s3_setup, tmpdir):
    file_to_get = "README.md"
    bucket_name = "test-bucket"

    s3_setup.fput_object(
        bucket_name=bucket_name, object_name=file_to_get, file_path=os.path.join(ELYRA_ROOT_DIR, file_to_get)
    )

    with tmpdir.as_cwd():
        op = _get_operation_instance(monkeypatch, s3_setup)

        op.get_file_from_object_storage(file_to_get)
        assert os.path.isfile(file_to_get)
        assert _fileChecksum(file_to_get) == _fileChecksum(os.path.join(ELYRA_ROOT_DIR, file_to_get))


def test_fail_get_file_object_store(monkeypatch, s3_setup, tmpdir):
    file_to_get = "test-file.txt"

    with tmpdir.as_cwd():
        with pytest.raises(minio.error.S3Error) as exc_info:
            op = _get_operation_instance(monkeypatch, s3_setup)
            op.get_file_from_object_storage(file_to_get=file_to_get)
        assert exc_info.value.code == "NoSuchKey"


def test_put_file_object_store(monkeypatch, s3_setup, tmpdir):
    bucket_name = "test-bucket"
    file_to_put = "LICENSE"

    op = _get_operation_instance(monkeypatch, s3_setup)
    op.put_file_to_object_storage(object_name=file_to_put, file_to_upload=os.path.join(ELYRA_ROOT_DIR, file_to_put))

    with tmpdir.as_cwd():
        s3_setup.fget_object(bucket_name, file_to_put, file_to_put)
        assert os.path.isfile(file_to_put)
        assert _fileChecksum(file_to_put) == _fileChecksum(os.path.join(ELYRA_ROOT_DIR, file_to_put))


def test_fail_invalid_filename_put_file_object_store(monkeypatch, s3_setup):
    file_to_put = "LICENSE_NOT_HERE"

    with pytest.raises(FileNotFoundError):
        op = _get_operation_instance(monkeypatch, s3_setup)
        op.put_file_to_object_storage(file_to_upload=file_to_put)


def test_fail_bucket_put_file_object_store(monkeypatch, s3_setup):
    bucket_name = "test-bucket-not-exist"
    file_to_put = "LICENSE"

    with pytest.raises(minio.error.S3Error) as exc_info:
        op = _get_operation_instance(monkeypatch, s3_setup)
        monkeypatch.setattr(op, "cos_bucket", bucket_name)
        op.put_file_to_object_storage(file_to_upload=os.path.join(ELYRA_ROOT_DIR, file_to_put))
    assert exc_info.value.code == "NoSuchBucket"


def test_find_best_kernel_nb(tmpdir):
    source_nb_file = os.path.join(RESOURCES_DIR, "test-notebookA.ipynb")
    nb_file = os.path.join(tmpdir, "test-notebookA.ipynb")

    # "Copy" nb file to destination - this test does not update the kernel or language.
    nb = nbformat.read(source_nb_file, 4)
    nbformat.write(nb, nb_file)

    with tmpdir.as_cwd():
        kernel_name = bootstrapper.NotebookFileOp.find_best_kernel(nb_file)
        assert kernel_name == nb.metadata.kernelspec["name"]


def test_find_best_kernel_lang(tmpdir, caplog):
    caplog.set_level(logging.INFO)
    source_nb_file = os.path.join(RESOURCES_DIR, "test-notebookA.ipynb")
    nb_file = os.path.join(tmpdir, "test-notebookA.ipynb")

    # "Copy" nb file to destination after updating the kernel name - forcing a language match
    nb = nbformat.read(source_nb_file, 4)
    nb.metadata.kernelspec["name"] = "test-kernel"
    nb.metadata.kernelspec["language"] = "PYTHON"  # test case-insensitivity
    nbformat.write(nb, nb_file)

    with tmpdir.as_cwd():
        kernel_name = bootstrapper.NotebookFileOp.find_best_kernel(nb_file)
        assert kernel_name == "python3"
        assert len(caplog.records) == 1
        assert caplog.records[0].message.startswith("Matched kernel by language (PYTHON)")


def test_find_best_kernel_nomatch(tmpdir, caplog):
    source_nb_file = os.path.join(RESOURCES_DIR, "test-notebookA.ipynb")
    nb_file = os.path.join(tmpdir, "test-notebookA.ipynb")

    # "Copy" nb file to destination after updating the kernel name and language - forcing use of updated name
    nb = nbformat.read(source_nb_file, 4)
    nb.metadata.kernelspec["name"] = "test-kernel"
    nb.metadata.kernelspec["language"] = "test-language"
    nbformat.write(nb, nb_file)

    with tmpdir.as_cwd():
        kernel_name = bootstrapper.NotebookFileOp.find_best_kernel(nb_file)
        assert kernel_name == "test-kernel"
        assert len(caplog.records) == 1
        assert caplog.records[0].message.startswith("Reverting back to missing notebook kernel 'test-kernel'")


def test_parse_arguments():
    test_args = [
        "-e",
        "http://test.me.now",
        "-d",
        "test-directory",
        "-t",
        "test-archive.tgz",
        "-f",
        "test-notebook.ipynb",
        "-b",
        "test-bucket",
        "-p",
        "/tmp/lib",
        "-n",
        "test-pipeline",
    ]
    args_dict = bootstrapper.OpUtil.parse_arguments(test_args)

    assert args_dict["cos-endpoint"] == "http://test.me.now"
    assert args_dict["cos-directory"] == "test-directory"
    assert args_dict["cos-dependencies-archive"] == "test-archive.tgz"
    assert args_dict["cos-bucket"] == "test-bucket"
    assert args_dict["filepath"] == "test-notebook.ipynb"
    assert args_dict["user-volume-path"] == "/tmp/lib"
    assert args_dict["pipeline-name"] == "test-pipeline"
    assert not args_dict["inputs"]
    assert not args_dict["outputs"]


def test_fail_missing_notebook_parse_arguments():
    test_args = ["-e", "http://test.me.now", "-d", "test-directory", "-t", "test-archive.tgz", "-b", "test-bucket"]
    with pytest.raises(SystemExit):
        bootstrapper.OpUtil.parse_arguments(test_args)


def test_fail_missing_endpoint_parse_arguments():
    test_args = ["-d", "test-directory", "-t", "test-archive.tgz", "-f", "test-notebook.ipynb", "-b", "test-bucket"]
    with pytest.raises(SystemExit):
        bootstrapper.OpUtil.parse_arguments(test_args)


def test_fail_missing_archive_parse_arguments():
    test_args = ["-e", "http://test.me.now", "-d", "test-directory", "-f", "test-notebook.ipynb", "-b", "test-bucket"]
    with pytest.raises(SystemExit):
        bootstrapper.OpUtil.parse_arguments(test_args)


def test_fail_missing_bucket_parse_arguments():
    test_args = [
        "-e",
        "http://test.me.now",
        "-d",
        "test-directory",
        "-t",
        "test-archive.tgz",
        "-f",
        "test-notebook.ipynb",
    ]
    with pytest.raises(SystemExit):
        bootstrapper.OpUtil.parse_arguments(test_args)


def test_fail_missing_directory_parse_arguments():
    test_args = ["-e", "http://test.me.now", "-t", "test-archive.tgz", "-f", "test-notebook.ipynb", "-b", "test-bucket"]
    with pytest.raises(SystemExit):
        bootstrapper.OpUtil.parse_arguments(test_args)


def test_requirements_file(monkeypatch, tmpdir, caplog):
    elyra_requirements_file = Path(__file__).parent / "resources/test-requirements-elyra.txt"
    elyra_correct_number_of_packages = 18
    elyra_packages = bootstrapper.OpUtil.package_list_to_dict(str(elyra_requirements_file))
    assert len(elyra_packages) == elyra_correct_number_of_packages

    current_requirements_file = Path(__file__).parent / "resources/test-requirements-current.txt"
    current_correct_number_of_packages = 15
    current_packages = bootstrapper.OpUtil.package_list_to_dict(str(current_requirements_file))
    assert len(current_packages) == current_correct_number_of_packages

    mocked_package_list_to_dict = mock.Mock(side_effect=[elyra_packages, current_packages])
    monkeypatch.setattr(bootstrapper.OpUtil, "package_list_to_dict", mocked_package_list_to_dict)

    mocked_subprocess_run = mock.Mock(return_value="default")
    monkeypatch.setattr(subprocess, "run", mocked_subprocess_run)

    bootstrapper.OpUtil.package_install(user_volume_path=str(tmpdir))
    assert "WARNING: Source package 'jupyter-client' found already installed as an editable package" in caplog.text
    assert "WARNING: Source package 'requests' found already installed as an editable package" in caplog.text
    assert "WARNING: Source package 'tornado' found already installed from git" in caplog.text


def test_fail_requirements_file_bad_delimiter():
    bad_requirements_file = Path(__file__).parent / "resources/test-bad-requirements-elyra.txt"
    with open(bad_requirements_file, "r") as f:
        file_content = f.readlines()
    valid_package_list = [
        line.strip("\n").split("==")[0] for line in file_content if not line.startswith("#") and "==" in line
    ]

    package_dict = bootstrapper.OpUtil.package_list_to_dict(bad_requirements_file)
    assert valid_package_list == list(package_dict.keys())


def _fileChecksum(filename):
    hasher = hashlib.sha256()

    with open(filename, "rb") as afile:
        buf = afile.read(65536)
        while len(buf) > 0:
            hasher.update(buf)
            buf = afile.read(65536)
    checksum = hasher.hexdigest()
    return checksum



#
# Copyright 2018-2023 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from datetime import datetime
from enum import Enum
from enum import unique
import hashlib
import importlib
import json
import os
from pathlib import Path
import re
import string
import sys
import tempfile
import time
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from urllib.parse import urlsplit

from autopep8 import fix_code
from jinja2 import Environment
from jinja2 import PackageLoader
from kfp import Client as ArgoClient
from kfp import compiler as kfp_argo_compiler
from kfp import components as components
from kubernetes import client as k8s_client
from traitlets import default
from traitlets import Unicode

RUN_ID_PLACEHOLDER = "{{workflow.uid}}"

from elyra._version import __version__
from elyra.metadata.schemaspaces import RuntimeImages
from elyra.metadata.schemaspaces import Runtimes
from elyra.pipeline import pipeline_constants
from elyra.pipeline.component_catalog import ComponentCache
from elyra.pipeline.kfp.kfp_authentication import AuthenticationError
from elyra.pipeline.kfp.kfp_authentication import KFPAuthenticator
from elyra.pipeline.kfp.kfp_properties import KfpPipelineParameter
from elyra.pipeline.kfp.PipelineConf import PipelineConf
from elyra.pipeline.pipeline import Operation
from elyra.pipeline.pipeline import Pipeline
from elyra.pipeline.processor import PipelineProcessor
from elyra.pipeline.processor import RuntimePipelineProcessor
from elyra.pipeline.processor import RuntimePipelineProcessorResponse
from elyra.pipeline.properties import CustomSharedMemorySize
from elyra.pipeline.properties import DisableNodeCaching
from elyra.pipeline.properties import ElyraProperty
from elyra.pipeline.properties import ElyraPropertyList
from elyra.pipeline.properties import KubernetesAnnotation
from elyra.pipeline.properties import KubernetesLabel
from elyra.pipeline.properties import KubernetesSecret
from elyra.pipeline.properties import KubernetesToleration
from elyra.pipeline.properties import PipelineParameter
from elyra.pipeline.properties import VolumeMount
from elyra.pipeline.runtime_type import RuntimeProcessorType
from elyra.util.cos import join_paths
from elyra.util.kubernetes import sanitize_label_value
from elyra.util.path import get_absolute_path


@unique
class WorkflowEngineType(Enum):
    """
    Identifies Kubeflow Pipelines workflow engines that this
    processor supports.
    """

    ARGO = "argo"
    TEKTON = "tekton"

    @staticmethod
    def get_instance_by_value(value: str) -> "WorkflowEngineType":
        """
        Produces an WorkflowEngineType enum instance if the provided value
        identifies a supported workflow engine type.
        Raises KeyError if value is not a support workflow engine type.
        """
        if value:
            for instance in WorkflowEngineType.__members__.values():
                if instance.value == value.lower():
                    return instance
        raise KeyError(f"'{value}'")


# Externalize these constants to make them available to the code gen tests
CRIO_VOL_DEF_NAME = "workspace"
CRIO_VOL_DEF_SIZE = "20Gi"
CRIO_VOL_DEF_MEDIUM = ""
CRIO_VOL_MOUNT_PATH = "/opt/app-root/src"
CRIO_VOL_WORKDIR_PATH = f"{CRIO_VOL_MOUNT_PATH}/jupyter-work-dir"
CRIO_VOL_PYTHON_PATH = f"{CRIO_VOL_WORKDIR_PATH}/python3"


class KfpPipelineProcessor(RuntimePipelineProcessor):
    _type = RuntimeProcessorType.KUBEFLOW_PIPELINES
    _name = "kfp"

    # Provide users with the ability to identify a writable directory in the
    # running container where the notebook | script is executed. The location
    # must exist and be known before the container is started.
    # Defaults to `/tmp`
    WCD = os.getenv("ELYRA_WRITABLE_CONTAINER_DIR", "/tmp").strip().rstrip("/")

    # Set the method for passing parameters to notebook and scripts
    # Only one value is currently supported ("env", which passes
    # parameters as environment variables)
    parameter_pass_default = "env"
    parameter_pass_method_env = "ELYRA_PARAMETER_PASS_METHOD"
    parameter_pass_method = Unicode(
        parameter_pass_default,
        help="""Sets the method to be used to pass pipeline parameters
                     to notebooks and scripts. Can be one of: ["env"].
                     (default=env). (ELYRA_PARAMETER_PASS_METHOD env var)""",
    ).tag(config=True)

    @default("parameter_pass_method")
    def parameter_pass_method_default(self):
        parameter_pass_default = KfpPipelineProcessor.parameter_pass_default
        try:
            parameter_pass_default = os.getenv(self.parameter_pass_method_env, parameter_pass_default)
        except ValueError:
            self.log.info(
                f"Unable to parse environmental variable {self.parameter_pass_method_env}, "
                f"using the default value of {self.parameter_pass_default}"
            )
        return parameter_pass_default

    def process(self, pipeline):
        """
        Runs a pipeline on Kubeflow Pipelines

        Each time a pipeline is processed, a new version
        is uploaded and run under the same experiment name.
        """

        timestamp = datetime.now().strftime("%m%d%H%M%S")

        ################
        # Runtime Configs
        ################
        runtime_configuration = self._get_metadata_configuration(
            schemaspace=Runtimes.RUNTIMES_SCHEMASPACE_ID, name=pipeline.runtime_config
        )

        # unpack Kubeflow Pipelines configs
        api_endpoint = runtime_configuration.metadata["api_endpoint"].rstrip("/")
        public_api_endpoint = runtime_configuration.metadata.get("public_api_endpoint", api_endpoint)
        api_username = runtime_configuration.metadata.get("api_username")
        api_password = runtime_configuration.metadata.get("api_password")
        user_namespace = runtime_configuration.metadata.get("user_namespace")
        workflow_engine = WorkflowEngineType.get_instance_by_value(runtime_configuration.metadata.get("engine", "argo"))

        # unpack Cloud Object Storage configs
        cos_endpoint = runtime_configuration.metadata["cos_endpoint"]
        cos_public_endpoint = runtime_configuration.metadata.get("public_cos_endpoint", cos_endpoint)
        cos_bucket = runtime_configuration.metadata["cos_bucket"]

        # Determine which provider to use to authenticate with Kubeflow
        auth_type = runtime_configuration.metadata.get("auth_type")

        try:
            auth_info = KFPAuthenticator().authenticate(
                api_endpoint,
                auth_type_str=auth_type,
                runtime_config_name=pipeline.runtime_config,
                auth_parm_1=api_username,
                auth_parm_2=api_password,
            )
            self.log.debug(f"Authenticator returned {auth_info}")
        except AuthenticationError as ae:
            if ae.get_request_history() is not None:
                self.log.info("An authentication error was raised. Diagnostic information follows.")
                self.log.info(ae.request_history_to_string())
            raise RuntimeError(f"Kubeflow authentication failed: {ae}")

        #############
        # Create Kubeflow Client
        #############
        try:
            client = ArgoClient(
                host=api_endpoint,
                cookies=auth_info.get("cookies", None),
                credentials=auth_info.get("credentials", None),
                existing_token=auth_info.get("existing_token", None),
                namespace=user_namespace,
                ssl_ca_cert=auth_info.get("ssl_ca_cert", None),
            )
        except Exception as ex:
            # a common cause of these errors is forgetting to include `/pipeline` or including it with an 's'
            api_endpoint_obj = urlsplit(api_endpoint)
            if api_endpoint_obj.path != "/pipeline":
                api_endpoint_tip = api_endpoint_obj._replace(path="/pipeline").geturl()
                tip_string = (
                    f" - [TIP: did you mean to set '{api_endpoint_tip}' as the endpoint, "
                    f"take care not to include 's' at end]"
                )
            else:
                tip_string = ""

            raise RuntimeError(
                f"Failed to initialize `kfp.Client()` against: '{api_endpoint}' - "
                f"Check Kubeflow Pipelines runtime configuration: '{pipeline.runtime_config}'"
                f"{tip_string}"
            ) from ex

        #############
        # Verify Namespace
        #############
        try:
            client.list_experiments(namespace=user_namespace, page_size=1)
        except Exception as ex:
            if user_namespace:
                tip_string = f"[TIP: ensure namespace '{user_namespace}' is correct]"
            else:
                tip_string = "[TIP: you probably need to set a namespace]"

            raise RuntimeError(
                f"Failed to `kfp.Client().list_experiments()` against: '{api_endpoint}' - "
                f"Check Kubeflow Pipelines runtime configuration: '{pipeline.runtime_config}' - "
                f"{tip_string}"
            ) from ex

        #############
        # Pipeline Metadata none - inherited
        #############
        # generate a pipeline name
        pipeline_name = pipeline.name

        # generate a pipeline description
        pipeline_description = pipeline.description
        if pipeline_description is None:
            pipeline_description = f"Created with Elyra {__version__} pipeline editor using `{pipeline.source}`."

        #############
        # Submit & Run the Pipeline
        #############
        self.log_pipeline_info(pipeline_name, "submitting pipeline")

        with tempfile.TemporaryDirectory() as temp_dir:
            self.log.debug(f"Created temporary directory at: {temp_dir}")
            pipeline_path = os.path.join(temp_dir, f"{pipeline_name}.yaml")

            #############
            # Get Pipeline ID
            #############
            try:
                # get the kubeflow pipeline id (returns None if not found, otherwise the ID of the pipeline)
                pipeline_id = client.get_pipeline_id(pipeline_name)

                # calculate what "pipeline version" name to use
                if pipeline_id is None:
                    # the first "pipeline version" name must be the pipeline name
                    pipeline_version_name = pipeline_name
                else:
                    # generate a unique name for a new "pipeline version" by appending the current timestamp
                    pipeline_version_name = f"{pipeline_name}-{timestamp}"

            except Exception as ex:
                raise RuntimeError(
                    f"Failed to get ID of Kubeflow pipeline: '{pipeline_name}' - "
                    f"Check Kubeflow Pipelines runtime configuration: '{pipeline.runtime_config}'"
                ) from ex

            #############
            # Compile the Pipeline
            #############
            try:
                t0 = time.time()

                # generate a name for the experiment (lowercase because experiments are case intensive)
                experiment_name = pipeline_name.lower()

                # Create an instance id that will be used to store
                # the pipelines' dependencies, if applicable
                pipeline_instance_id = f"{pipeline_name}-{timestamp}"

                # Generate Python DSL from workflow
                pipeline_dsl = self._generate_pipeline_dsl(
                    pipeline=pipeline,
                    pipeline_name=pipeline_name,
                    pipeline_instance_id=pipeline_instance_id,
                    workflow_engine=workflow_engine,
                )

                # Collect pipeline configuration information
                pipeline_conf = self._generate_pipeline_conf(pipeline=pipeline)

                # Compile the Python DSL, producing the input for the upload to
                # Kubeflow Pipelines
                self._compile_pipeline_dsl(pipeline_dsl, workflow_engine, pipeline_path, pipeline_conf)

            except RuntimeError:
                raise
            except Exception as ex:
                raise RuntimeError(
                    f"Error compiling pipeline '{pipeline_name}' with engine '{workflow_engine.value}'."
                ) from ex

            self.log_pipeline_info(pipeline_name, "pipeline compiled", duration=time.time() - t0)

            #############
            # Upload Pipeline Version
            #############
            try:
                t0 = time.time()

                # CASE 1: pipeline needs to be created
                if pipeline_id is None:
                    # create new pipeline (and initial "pipeline version")
                    kfp_pipeline = client.upload_pipeline(
                        pipeline_package_path=pipeline_path,
                        pipeline_name=pipeline_name,
                        description=pipeline_description,
                    )

                    # extract the ID of the pipeline we created
                    pipeline_id = kfp_pipeline.pipeline_id

                    # the initial "pipeline version" has the same id as the pipeline itself
                    version_details = client.list_pipeline_versions(pipeline_id=pipeline_id)
                    version_list = version_details.pipeline_versions
                    if isinstance(version_list, list):
                        version_id = version_list[0].pipeline_version_id
                    else:
                        version_id = None
                # CASE 2: pipeline already exists
                else:
                    # upload the "pipeline version"
                    kfp_pipeline = client.upload_pipeline_version(
                        pipeline_package_path=pipeline_path,
                        pipeline_version_name=pipeline_version_name,
                        pipeline_id=pipeline_id,
                    )

                    # extract the id of the "pipeline version" that was created
                    version_id = kfp_pipeline.pipeline_version_id

            except Exception as ex:
                # a common cause of these errors is forgetting to include `/pipeline` or including it with an 's'
                api_endpoint_obj = urlsplit(api_endpoint)
                if api_endpoint_obj.path != "/pipeline":
                    api_endpoint_tip = api_endpoint_obj._replace(path="/pipeline").geturl()
                    tip_string = (
                        f" - [TIP: did you mean to set '{api_endpoint_tip}' as the endpoint, "
                        f"take care not to include 's' at end]"
                    )
                else:
                    tip_string = ""

                raise RuntimeError(
                    f"Failed to upload Kubeflow pipeline '{pipeline_name}' - "
                    f"Check Kubeflow Pipelines runtime configuration: '{pipeline.runtime_config}'"
                    f"{tip_string}"
                ) from ex

            self.log_pipeline_info(pipeline_name, "pipeline uploaded", duration=time.time() - t0)

            #############
            # Create Experiment
            #############
            try:
                t0 = time.time()

                # create a new experiment (if already exists, this a no-op)
                experiment = client.create_experiment(name=experiment_name, namespace=user_namespace)

            except Exception as ex:
                raise RuntimeError(
                    f"Failed to create Kubeflow experiment: '{experiment_name}' - "
                    f"Check Kubeflow Pipelines runtime configuration: '{pipeline.runtime_config}'"
                ) from ex

            self.log_pipeline_info(pipeline_name, "created experiment", duration=time.time() - t0)

            #############
            # Create Pipeline Run
            #############
            try:
                t0 = time.time()

                # generate name for the pipeline run
                job_name = pipeline_instance_id

                # create pipeline run (or specified pipeline version)
                run = client.run_pipeline(
                    experiment_id=experiment.experiment_id,
                    job_name=job_name,
                    pipeline_id=pipeline_id,
                    version_id=version_id,
                )

            except Exception as ex:
                raise RuntimeError(
                    f"Failed to create Kubeflow pipeline run: '{job_name}' - "
                    f"Check Kubeflow Pipelines runtime configuration: '{pipeline.runtime_config}'"
                ) from ex

            if run is None:
                # client.run_pipeline seemed to have encountered an issue
                # but didn't raise an exception
                raise RuntimeError(
                    f"Failed to create Kubeflow pipeline run: '{job_name}' - "
                    f"Check Kubeflow Pipelines runtime configuration: '{pipeline.runtime_config}'"
                )

            self.log_pipeline_info(
                pipeline_name,
                f"pipeline submitted: {public_api_endpoint}/{experiment.experiment_id}/runs/{run.run_id}",
                duration=time.time() - t0,
            )

        if pipeline.contains_generic_operations():
            object_storage_url = f"{cos_public_endpoint}"
            os_path = join_paths(
                pipeline.pipeline_properties.get(pipeline_constants.COS_OBJECT_PREFIX), pipeline_instance_id
            )
            object_storage_path = f"/{cos_bucket}/{os_path}"
        else:
            object_storage_url = None
            object_storage_path = None

        return KfpPipelineProcessorResponse(
            run_id=run.run_id,
            run_url=f"{public_api_endpoint}/{experiment.experiment_id}/runs/{run.run_id}",
            object_storage_url=object_storage_url,
            object_storage_path=object_storage_path,
        )

    def export(
        self, pipeline: Pipeline, pipeline_export_format: str, pipeline_export_path: str, overwrite: bool
    ) -> str:
        """
        Export pipeline to the specified format and store the output
        in the specified file.

        :param pipeline: The pipeline to be exported
        :type pipeline: Pipeline
        :param pipeline_export_format: "py" for KFP Python DSL or "yaml" for YAML
        :type pipeline_export_format: str
        :param pipeline_export_path: name and location of exported file
        :type pipeline_export_path: str
        :param overwrite: If false, export raises an error if the output file exists.
        :type overwrite: bool
        :raises ValueError: raised if a parameter is invalid
        :raises RuntimeError: an error occurred during export
        :return: location of the exported file
        :rtype: str
        """
        # Verify that the processor supports the given export format
        self._verify_export_format(pipeline_export_format)
        t0_all = time.time()
        timestamp = datetime.now().strftime("%m%d%H%M%S")
        pipeline_name = pipeline.name
        # Create an instance id that will be used to store
        # the pipelines' dependencies, if applicable
        pipeline_instance_id = f"{pipeline_name}-{timestamp}"

        # Since pipeline_export_path may be relative to the notebook directory, ensure
        # we're using its absolute form.
        absolute_pipeline_export_path = get_absolute_path(self.root_dir, pipeline_export_path)

        runtime_configuration = self._get_metadata_configuration(
            schemaspace=Runtimes.RUNTIMES_SCHEMASPACE_ID, name=pipeline.runtime_config
        )

        workflow_engine = WorkflowEngineType.get_instance_by_value(runtime_configuration.metadata.get("engine", "argo"))

        if Path(absolute_pipeline_export_path).exists() and not overwrite:
            raise ValueError("File " + absolute_pipeline_export_path + " already exists.")

        self.log_pipeline_info(pipeline_name, f"Exporting pipeline as a .{pipeline_export_format} file")
        try:
            # Generate Python DSL
            pipeline_dsl = self._generate_pipeline_dsl(
                pipeline=pipeline,
                pipeline_name=pipeline_name,
                pipeline_instance_id=pipeline_instance_id,
                workflow_engine=workflow_engine,
            )

            if pipeline_export_format == "py":
                # Write Python DSL to file
                with open(absolute_pipeline_export_path, "w") as dsl_output:
                    dsl_output.write(pipeline_dsl)
            else:
                # Generate pipeline configuration
                pipeline_conf = self._generate_pipeline_conf(pipeline=pipeline)
                #
                # Export pipeline as static configuration file (YAML formatted)
                # by invoking the compiler for the selected engine
                self._compile_pipeline_dsl(pipeline_dsl, workflow_engine, absolute_pipeline_export_path, pipeline_conf)
        except RuntimeError:
            raise
        except Exception as ex:
            if ex.__cause__:
                raise RuntimeError(str(ex)) from ex
            raise RuntimeError(
                f"Error pre-processing pipeline '{pipeline_name}' for export to '{absolute_pipeline_export_path}'",
                str(ex),
            ) from ex

        self.log_pipeline_info(
            pipeline_name, f"pipeline exported to '{pipeline_export_path}'", duration=(time.time() - t0_all)
        )

        return pipeline_export_path  # Return the input value, not its absolute form

    def _collect_envs(self, operation: Operation, **kwargs) -> Dict:
        """
        Amends envs collected from superclass with those pertaining to this subclass

        :return: dictionary containing environment name/value pairs
        """
        envs = super()._collect_envs(operation, **kwargs)
        # Only Unix-style path spec is supported.
        envs["ELYRA_WRITABLE_CONTAINER_DIR"] = self.WCD
        return envs

    def _generate_pipeline_dsl(
        self,
        pipeline: Pipeline,
        pipeline_name: str,
        workflow_engine: WorkflowEngineType,
        pipeline_version: str = "",
        experiment_name: str = "",
        pipeline_instance_id: str = None,
        code_generation_options: Optional[Dict[str, Any]] = None,
    ) -> str:
        """
        Generate Python DSL for Kubeflow Pipelines v1
        """
        if not code_generation_options:
            code_generation_options = {}

        # Load Kubeflow Pipelines Python DSL template
        loader = PackageLoader("elyra", "templates/kubeflow/v2")
        template_env = Environment(loader=loader)
        # Add filter that produces a Python-safe variable name
        template_env.filters["python_safe"] = lambda x: re.sub(r"[" + re.escape(string.punctuation) + "\\s]", "_", x)
        # Add filter that escapes the " character in strings
        template_env.filters["string_delimiter_safe"] = lambda string: re.sub('"', '\\"', string)
        # Add filter that converts a value to a python variable value (e.g. puts quotes around strings)
        template_env.filters["param_val_to_python_var"] = lambda p: (
            "None" if p.value is None else (f'"{p.value}"' if p.input_type.base_type == "String" else p.value)
        )
        template = template_env.get_template("python_dsl_template.jinja2")

        # Convert pipeline into workflow tasks
        workflow_tasks = self._generate_workflow_tasks(
            pipeline,
            pipeline_name,
            workflow_engine,
            pipeline_instance_id=pipeline_instance_id,
            pipeline_version=pipeline_version,
            experiment_name=experiment_name,
        )

        # Gather unique component definitions from workflow task list.
        unique_component_definitions = {}
        for key, operation in workflow_tasks.items():
            unique_component_definitions[operation["component_definition_hash"]] = operation["component_definition"]

        pipeline_parameters: List[PipelineParameter] = pipeline.parameters
        if code_generation_options.get("render_all_parameters") is not True:
            # Gather the list of parameter names referenced by the pipeline's tasks
            referenced_parameter_names: Set[str] = set()
            for task in workflow_tasks.values():
                for task_details in task["task_inputs"].values():
                    parameter_reference = task_details.get("pipeline_parameter_reference")
                    if parameter_reference:
                        referenced_parameter_names.add(parameter_reference)

            # Reduce set of pipeline parameters to those referenced by pipeline tasks
            for parameter in pipeline.parameters:
                if parameter.name not in referenced_parameter_names:
                    pipeline_parameters.remove(parameter)

        # render the Kubeflow Pipelines Python DSL template
        pipeline_dsl = template.render(
            elyra_version=__version__,
            pipeline_name=pipeline_name,
            pipeline_description=pipeline.description,
            pipeline_parameters=pipeline_parameters,
            workflow_tasks=workflow_tasks,
            component_definitions=unique_component_definitions,
            workflow_engine=workflow_engine.value,
        )

        # Prettify generated Python DSL
        # Defer the import to postpone logger messages: https://github.com/psf/black/issues/2058
        import black

        try:
            pipeline_dsl = black.format_str(fix_code(pipeline_dsl), mode=black.FileMode())
        except Exception:
            # if an error was encountered log the generated DSL for troubleshooting
            self.log.error("Error post-processing generated Python DSL:")
            self.log.error(pipeline_dsl)
            raise

        return pipeline_dsl

    def _compile_pipeline_dsl(
        self, dsl: str, workflow_engine: WorkflowEngineType, output_file: str, pipeline_conf: PipelineConf
    ) -> None:
        """
        Compile Python DSL using the compiler for the specified workflow_engine.

        :param dsl: the Python DSL to be compiled
        :type dsl: str
        :param workflow_engine: Compiler to be used
        :type workflow_engine: str
        :param output_file: output file name
        :type output_file: str
        :param pipeline_conf: Pipeline configuration to apply
        :type pipeline_conf: PipelineConf
        :raises RuntimeError: raised when a fatal error is encountered
        """

        with tempfile.TemporaryDirectory() as temp_dir:
            module_name = "generated_dsl"
            try:
                # Add temporary directory to Python module search path.
                sys.path.insert(0, temp_dir)
                # Save DSL in temporary file so we can import it as a module.
                dsl_file = Path(temp_dir) / f"{module_name}.py"
                with open(dsl_file, "w") as dsl_output:
                    dsl_output.write(dsl)
                # Load DSL by importing the "generated_dsl" module.
                mod = importlib.import_module(module_name)
                # If this module was previously imported it won't reflect
                # changes that might be in the DSL we are about to compile.
                # Force a module re-load to pick up any changes.
                mod = importlib.reload(mod)
                # Obtain handle to pipeline function, which is named
                # in the generated Python DSL "generated_pipeline"
                pipeline_function = getattr(mod, "generated_pipeline")
                # compile the DSL
                kfp_argo_compiler.Compiler().compile(pipeline_function, output_file)
            except Exception as ex:
                raise RuntimeError(
                    f"Failed to compile pipeline with workflow_engine '{workflow_engine.value}' to '{output_file}'"
                ) from ex
            finally:
                # remove temporary directory from Python module search path
                del sys.path[0]
                # remove module entry; it's no longer needed now that it was
                # processed by the Kubeflow Pipelines compiler
                sys.modules.pop(module_name, None)

    def _generate_workflow_tasks(
        self,
        pipeline: Pipeline,
        pipeline_name: str,
        workflow_engine: WorkflowEngineType,
        pipeline_version: str = "",
        experiment_name: str = "",
        pipeline_instance_id: str = None,
        export: bool = False,
    ) -> Dict[str, Dict]:
        """
        Produce the workflow tasks that implement the pipeline nodes. The output is
        a dictionary containing task ids as keys and task definitions as values.
        """

        pipeline_instance_id = pipeline_instance_id or pipeline_name

        self.log_pipeline_info(
            pipeline_name,
            "Processing pipeline",
        )
        t0_all = time.time()

        # Sort operations based on dependency graph (topological order)
        sorted_operations = PipelineProcessor._sort_operations(pipeline.operations)

        if any(operation.is_generic for operation in sorted_operations):
            # The pipeline contains atleast one node that is implemented
            # using a generic component: collect and verify relevant information
            runtime_configuration = self._get_metadata_configuration(
                schemaspace=Runtimes.RUNTIMES_SCHEMASPACE_ID, name=pipeline.runtime_config
            )
            # - verify that cloud storage can be accessed
            self._verify_cos_connectivity(runtime_configuration)
            # - collect runtime configuration information
            cos_username = runtime_configuration.metadata.get("cos_username")
            cos_password = runtime_configuration.metadata.get("cos_password")
            cos_secret = runtime_configuration.metadata.get("cos_secret")
            cos_endpoint = runtime_configuration.metadata["cos_endpoint"]
            cos_bucket = runtime_configuration.metadata.get("cos_bucket")
            artifact_object_prefix = join_paths(
                pipeline.pipeline_properties.get(pipeline_constants.COS_OBJECT_PREFIX), pipeline_instance_id
            )
            # - load the generic component definition template
            template_env = Environment(loader=PackageLoader("elyra", "templates/kubeflow/v2"))
            generic_component_template = template_env.get_template("generic_component_definition_template.jinja2")
            # Add filter that escapes the " character in strings
            template_env.filters["string_delimiter_safe"] = lambda string: re.sub('"', '\\\\\\\\"', string)
            # Determine whether we are executing in a CRI-O runtime environment
            is_crio_runtime = os.getenv("CRIO_RUNTIME", "False").lower() == "true"

        # All previous operation outputs should be propagated throughout the pipeline.
        # In order to process this recursively, the current operation's inputs should be combined
        # from its parent's inputs (which, themselves are derived from the outputs of their parent)
        # and its parent's outputs.

        PipelineProcessor._propagate_operation_inputs_outputs(pipeline, sorted_operations)

        # Scrub all node labels of invalid characters
        for operation in sorted_operations:
            operation.name = re.sub("-+", "-", re.sub("[^-_0-9A-Za-z ]+", "-", operation.name)).lstrip("-").rstrip("-")

        # Generate unique names for all operations
        unique_names = {}
        for operation in sorted_operations:
            # Ensure operation name is unique
            new_name = operation.name
            while new_name in unique_names:
                new_name = f"{operation.name}_{unique_names[operation.name]}"
                unique_names[operation.name] += 1
            operation.name = new_name
            unique_names[operation.name] = 1

        # Create workflow task list that is used as input for the DSL code generator
        # from the sorted list of pipeline.pipeline.Operations
        workflow_tasks = {}
        for operation in sorted_operations:
            # Create workflow task, which Jinja2 uses to generate the task specific
            # source code.
            workflow_task = {
                "id": operation.id,
                "escaped_task_id": re.sub(r"[" + re.escape(string.punctuation) + "\\s]", "_", operation.id),
                "name": operation.name,
                "doc": operation.doc,
                "upstream_workflow_task_ids": operation.parent_operation_ids or [],
                "task_inputs": {},  # as defined in the component specification
                "task_outputs": {},  # as defined in the component specification
                "task_modifiers": {},  # attached volumes, resources, env variables, metadata, etc
            }

            # Add Elyra-owned properties (data volume mounts, kubernetes labels, etc)
            # to the task_modifiers property.
            for value in operation.elyra_props.values():
                if isinstance(value, (ElyraProperty, ElyraPropertyList)):
                    value.add_to_execution_object(
                        runtime_processor=self, execution_object=workflow_task["task_modifiers"]
                    )

            if operation.is_generic:
                # The task is implemented using a generic component
                workflow_task["uses_custom_component"] = False

                # Set parameters specified for this task and add each as a task input
                task_parameters = [param for param in pipeline.parameters if param.name in operation.parameters]
                workflow_task["task_inputs"] = {
                    param.name: {"pipeline_parameter_reference": param.name} for param in task_parameters
                }

                component_definition = generic_component_template.render(
                    container_image=operation.runtime_image,
                    task_parameters=task_parameters,
                    command_args=self._compose_container_command_args(
                        pipeline_name=pipeline_name,
                        cos_endpoint=cos_endpoint,
                        cos_bucket=cos_bucket,
                        cos_directory=artifact_object_prefix,
                        cos_dependencies_archive=self._get_dependency_archive_name(operation),
                        filename=operation.filename,
                        cos_inputs=operation.inputs,
                        cos_outputs=operation.outputs,
                        task_parameters=task_parameters,
                        is_crio_runtime=is_crio_runtime,
                    ),
                )
                workflow_task["component_definition"] = component_definition
                workflow_task["component_definition_hash"] = hashlib.sha256(component_definition.encode()).hexdigest()

                # attach environment variables
                workflow_task["task_modifiers"]["env_variables"] = self._collect_envs(
                    operation, cos_secret=cos_secret, cos_username=cos_username, cos_password=cos_password
                )

                # hack only: since we don't use the ContainerOp constructor anymore
                # we cannot use the file_outputs parameter to provide the information
                # https://www.kubeflow.org/docs/components/pipelines/v1/sdk/output-viewer/
                workflow_task["task_modifiers"]["special_output_files"] = {
                    "mlpipeline_ui_metadata": (Path(self.WCD) / "mlpipeline-ui-metadata.json").as_posix(),
                    "mlpipeline_metrics": (Path(self.WCD) / "mlpipeline-metrics.json").as_posix(),
                }

                # apply object storage Kubernetes secret, if one was provided
                if cos_secret and not export:
                    workflow_task["task_modifiers"]["object_storage_secret"] = cos_secret

                # apply container image pull policy, if one was specified
                for image_instance in self._get_metadata_configuration(RuntimeImages.RUNTIME_IMAGES_SCHEMASPACE_ID):
                    if image_instance.metadata["image_name"] == operation.runtime_image and image_instance.metadata.get(
                        "pull_policy"
                    ):
                        workflow_task["task_modifiers"]["image_pull_policy"] = image_instance.metadata["pull_policy"]
                        break

                # set resource constraints
                workflow_task["task_modifiers"]["cpu_request"] = operation.cpu
                workflow_task["task_modifiers"]["mem_request"] = {
                    "size": operation.memory,
                    "units": "G",
                }
                workflow_task["task_modifiers"]["cpu_limit"] = operation.cpu_limit
                workflow_task["task_modifiers"]["memory_limit"] = {
                    "size": operation.memory_limit,
                    "units": "G",
                }
                gpu_vendor = "nvidia.com/gpu"
                if operation.gpu_vendor:
                    gpu_vendor = operation.gpu_vendor
                workflow_task["task_modifiers"]["gpu_limit"] = {"size": operation.gpu, "vendor": gpu_vendor}

                if is_crio_runtime:
                    # Attach empty dir volume
                    workflow_task["task_modifiers"]["crio_runtime"] = {
                        "emptydir_volume_name": CRIO_VOL_DEF_NAME,
                        "emptydir_volume_size": CRIO_VOL_DEF_SIZE,
                        "emptydir_volume_medium": CRIO_VOL_DEF_MEDIUM,
                        "emptydir_mount_path": CRIO_VOL_MOUNT_PATH,
                    }
                    # Set Python module search path
                    workflow_task["task_modifiers"]["env_variables"]["PYTHONPATH"] = CRIO_VOL_PYTHON_PATH

                # Attach identifying metadata
                if workflow_task["task_modifiers"].get("pod_labels") is None:
                    workflow_task["task_modifiers"]["pod_labels"] = {}
                # Node type (a static type for this op)
                workflow_task["task_modifiers"]["pod_labels"]["elyra/node-type"] = sanitize_label_value(
                    "notebook-script"
                )
                # Pipeline name
                workflow_task["task_modifiers"]["pod_labels"]["elyra/pipeline-name"] = sanitize_label_value(
                    pipeline_name
                )
                # Pipeline version
                workflow_task["task_modifiers"]["pod_labels"]["elyra/pipeline-version"] = sanitize_label_value(
                    pipeline_version
                )
                # Experiment name
                workflow_task["task_modifiers"]["pod_labels"]["elyra/experiment-name"] = sanitize_label_value(
                    experiment_name
                )
                # Pipeline node name
                workflow_task["task_modifiers"]["pod_labels"]["elyra/node-name"] = sanitize_label_value(operation.name)

                # Add non-identifying metadata
                if workflow_task["task_modifiers"].get("pod_annotations") is None:
                    workflow_task["task_modifiers"]["pod_annotations"] = {}
                # Pipeline node file
                workflow_task["task_modifiers"]["pod_annotations"]["elyra/node-file-name"] = operation.filename

                # Identify the pipeline source, which can be a pipeline file (mypipeline.pipeline), a Python
                # script or notebook that was submitted
                if pipeline.source is not None:
                    workflow_task["task_modifiers"]["pod_annotations"]["elyra/pipeline-source"] = pipeline.source

                # Generate unique ELYRA_RUN_NAME value, which gets exposed as an environment
                # variable
                if workflow_engine == WorkflowEngineType.TEKTON:
                    # Value is derived from an existing annotation; use dummy value
                    workflow_task["task_modifiers"]["set_run_name"] = "dummy value"
                else:
                    # Use Kubeflow Pipelines provided RUN_ID_PLACEHOLDER as run name
                    workflow_task["task_modifiers"]["set_run_name"] = RUN_ID_PLACEHOLDER

                # Upload dependencies to cloud storage
                self._upload_dependencies_to_object_store(
                    runtime_configuration, pipeline_name, operation, prefix=artifact_object_prefix
                )

            else:
                # ----------------------------------------
                # The task is implemented using a custom component
                workflow_task["uses_custom_component"] = True

                # Retrieve component from cache
                component = ComponentCache.instance().get_component(self._type, operation.classifier)

                workflow_task["component_definition"] = component.definition
                workflow_task["component_definition_hash"] = hashlib.sha256(component.definition.encode()).hexdigest()

                # Identify task inputs and outputs using the component spec
                # If no data type was specified, string is assumed
                factory_function = components.load_component_from_text(component.definition)
                for input_key, input_value in (factory_function.component_spec.inputs or {}).items():
                    sanitized_input_name = self._sanitize_param_name(input_key)
                    workflow_task["task_inputs"][sanitized_input_name] = {
                        "value": None,
                        "task_output_reference": None,
                        "pipeline_parameter_reference": None,
                        "data_type": (input_value.type or "string").lower(),
                    }
                    # Determine whether the value needs to be rendered in quotes
                    # in the generated DSL code. For example "my name" (string), and 34 (integer).
                    workflow_task["task_inputs"][sanitized_input_name]["requires_quoted_rendering"] = workflow_task[
                        "task_inputs"
                    ][sanitized_input_name]["data_type"] not in [
                        "integer",
                        "float",
                        "bool",
                    ]

                for output_key, output_value in (factory_function.component_spec.outputs or {}).items():
                    workflow_task["task_outputs"][self._sanitize_param_name(output_key)] = {
                        "data_type": output_value.type,
                    }

                # Iterate over component properties and assign values to
                # task inputs and task add-ons
                for component_property in component.properties:
                    self.log.debug(
                        f"Processing component property '{component_property.name}' "
                        f"of type '{component_property.json_data_type}'"
                    )

                    if component_property.allowed_input_types == [None]:
                        # The property does not support inputs. Ignore
                        continue

                    sanitized_component_property_id = self._sanitize_param_name(component_property.ref)
                    if sanitized_component_property_id in workflow_task["task_inputs"]:
                        reference = workflow_task["task_inputs"][sanitized_component_property_id]
                    else:
                        workflow_task["task_modifiers"][sanitized_component_property_id] = {}
                        reference = workflow_task["task_modifiers"][sanitized_component_property_id]

                    # Get corresponding property's value from parsed pipeline
                    property_value_dict = operation.component_props.get(component_property.ref)
                    data_entry_type = property_value_dict.get("widget", None)  # one of: inputpath, file, raw data type
                    property_value = property_value_dict.get("value", None)
                    if data_entry_type == "inputpath":
                        # task input is the output of an upstream task
                        output_node_id = property_value["value"]  # parent node id
                        output_node_property_key = property_value["option"].replace("output_", "")  # parent property
                        reference["task_output_reference"] = {
                            "task_id": re.sub(r"[" + re.escape(string.punctuation) + "\\s]", "_", output_node_id),
                            "output_id": self._sanitize_param_name(output_node_property_key),
                        }
                    elif data_entry_type == "parameter":
                        # task input is the name of a pipeline parameter
                        param_name = property_value
                        if param_name is None or param_name not in pipeline.parameters.to_dict():
                            # Parameter name not found in list, fall back to using the raw default value
                            reference["value"] = component_property.value
                        else:
                            # Set pipeline parameter reference for this task
                            reference["pipeline_parameter_reference"] = param_name

                    else:
                        # task input is a raw value, either from file contents or entered manually
                        if data_entry_type == "file" and property_value:
                            # Read value from the specified file
                            absolute_path = get_absolute_path(self.root_dir, property_value)
                            with open(absolute_path, "r") as f:
                                property_value = f.read() if os.path.getsize(absolute_path) else None

                        # If the value is not found, assign it the default value assigned in parser
                        if property_value is None:
                            property_value = component_property.value

                        # Process the value according to its type, if necessary
                        if component_property.json_data_type == "object":
                            reference["value"] = self._process_dictionary_value(property_value)
                        elif component_property.json_data_type == "array":
                            reference["value"] = self._process_list_value(property_value)
                        else:
                            reference["value"] = property_value

            self.log.debug(f"Completed processing of task '{workflow_task['name']}':")
            self.log.debug(json.dumps(workflow_task, sort_keys=False, indent=4))

            # append task to task list
            workflow_tasks[workflow_task["id"]] = workflow_task

        # end of processing
        self.log_pipeline_info(pipeline_name, "Pipeline processed", duration=(time.time() - t0_all))
        return workflow_tasks

    def _generate_pipeline_conf(self, pipeline: Pipeline) -> PipelineConf:
        """
        Returns a KFP pipeline configuration for this pipeline, which can be empty.

        :return: https://kubeflow-pipelines.readthedocs.io/en/latest/source/kfp.dsl.html#kfp.dsl.PipelineConf
        :rtype: kfp.dsl import PipelineConf
        """

        self.log.debug("Generating pipeline configuration ...")
        pipeline_conf = PipelineConf()

        #
        # Gather input for container image pull secrets in support of private container image registries
        # https://kubeflow-pipelines.readthedocs.io/en/latest/source/kfp.dsl.html#kfp.dsl.PipelineConf.set_image_pull_secrets
        #
        # Retrieve all runtime image configurations
        runtime_image_configurations = self._get_metadata_configuration(RuntimeImages.RUNTIME_IMAGES_SCHEMASPACE_ID)
        # For each generic pipeline operation determine wether its runtime image
        # is protected by a pull secret
        container_image_pull_secret_names = []
        for operation in pipeline.operations.values():
            if operation.is_generic:
                for image_instance in runtime_image_configurations:
                    if image_instance.metadata["image_name"] == operation.runtime_image:
                        if image_instance.metadata.get("pull_secret"):
                            container_image_pull_secret_names.append(image_instance.metadata.get("pull_secret"))
                        break

        if len(container_image_pull_secret_names) > 0:
            # de-duplicate the pull secret name list, create Kubernetes resource
            # references and add them to the pipeline configuration
            container_image_pull_secrets = []
            for secret_name in list(set(container_image_pull_secret_names)):
                container_image_pull_secrets.append(k8s_client.V1ObjectReference(name=secret_name))
            pipeline_conf.set_image_pull_secrets(container_image_pull_secrets)
            self.log.debug(
                f"Added {len(container_image_pull_secrets)}" " image pull secret(s) to the pipeline configuration."
            )

        return pipeline_conf

    def _compose_container_command_args(
        self,
        pipeline_name: str,
        cos_endpoint: str,
        cos_bucket: str,
        cos_directory: str,
        cos_dependencies_archive: str,
        filename: str,
        cos_inputs: Optional[List[str]] = [],
        cos_outputs: Optional[List[str]] = [],
        task_parameters: Optional[List[PipelineParameter]] = None,
        is_crio_runtime: bool = False,
    ) -> List[str]:
        """
        Compose the container command arguments for a generic component, taking into
        account whether the container will run in a CRI-O environment.
        """

        # NOTE: The default organization and default branch are
        # meant for the opendatahub-io/Data-science-pipelines v2 use case
        # and should not be opened as PR against upstream Elyra
        elyra_github_org = os.getenv("ELYRA_GITHUB_ORG", "opendatahub-io")
        elyra_github_branch = os.getenv("ELYRA_GITHUB_BRANCH", "main" if "dev" in __version__ else f"v{__version__}")
        elyra_bootstrap_script_url = os.getenv(
            "ELYRA_BOOTSTRAP_SCRIPT_URL",
            f"https://raw.githubusercontent.com/{elyra_github_org}/elyra/{elyra_github_branch}/elyra/kfp/bootstrapper.py",  # noqa E501
        )

        elyra_requirements_url = os.getenv(
            "ELYRA_REQUIREMENTS_URL",
            f"https://raw.githubusercontent.com/{elyra_github_org}/"
            f"elyra/{elyra_github_branch}/etc/generic/requirements-elyra.txt",
        )

        if is_crio_runtime:
            container_work_dir = CRIO_VOL_WORKDIR_PATH
            container_python_path = CRIO_VOL_PYTHON_PATH
            python_pip_config_url = os.getenv(
                "ELYRA_PIP_CONFIG_URL",
                f"https://raw.githubusercontent.com/{elyra_github_org}/elyra/{elyra_github_branch}/etc/kfp/pip.conf",
            )
            python_user_lib_path_target = f"--target={CRIO_VOL_PYTHON_PATH}"
        else:
            container_work_dir = "./jupyter-work-dir"
            python_user_lib_path_target = ""

        common_curl_options = "--fail -H 'Cache-Control: no-cache'"

        # Verify utils files are present in the container
        # if not, then download them as per the provided URLs.
        file_base = os.getenv("ELYRA_FILE_BASE_PATH", "/opt/app-root/bin/utils")

        command_args = [
            f"mkdir -p {container_work_dir} && cd {container_work_dir}",
            f"[[ -e '{file_base}/bootstrapper.py' ]] && (echo 'bootstrapper.py file already exists'; cp '{file_base}/bootstrapper.py' .) || "  # noqa E501
            f"(echo 'Downloading {elyra_bootstrap_script_url}'; curl {common_curl_options} -L '{elyra_bootstrap_script_url}' --output bootstrapper.py)",  # noqa E501
            f"[[ -e '{file_base}/requirements-elyra.txt' ]] && (echo 'requirements-elyra.txt file already exists'; cp '{file_base}/requirements-elyra.txt' .) || "  # noqa E501
            f"(echo 'Downloading {elyra_requirements_url}'; curl {common_curl_options} -L '{elyra_requirements_url}' --output requirements-elyra.txt)",  # noqa E501
        ]

        if is_crio_runtime:
            command_args.append(
                f"mkdir {container_python_path} && cd {container_python_path} && "
                f"[[ -e '{file_base}/pip.conf' ]] && (echo 'pip.conf file already exists'; cp $file_base/pip.conf .) ||"
                f"(echo 'Downloading {python_pip_config_url}'; curl {common_curl_options} -L '{python_pip_config_url}' --output pip.conf)",  # noqa E501
                "&& cd .. && ",
            )

        bootstrapper_command = [
            f"python3 -m pip install {python_user_lib_path_target} packaging && "
            "python3 -m pip freeze > requirements-current.txt && "
            "python3 bootstrapper.py "
            f"--pipeline-name '{pipeline_name}' "
            f"--cos-endpoint '{cos_endpoint}' "
            f"--cos-bucket '{cos_bucket}' "
            f"--cos-directory '{cos_directory}' "
            f"--cos-dependencies-archive '{cos_dependencies_archive}' "
            f"--file '{filename}' "
        ]

        def list_to_string(item_list: List[str]) -> str:
            """
            Utility function that converts a list of strings to a string
            """
            # Inputs and Outputs separator character.  If updated,
            # same-named variable in bootstrapper.py must be updated!
            INOUT_SEPARATOR = ";"
            for item in item_list:
                if INOUT_SEPARATOR in item:
                    raise ValueError(f"Illegal character ({INOUT_SEPARATOR}) found in list item '{item}'.")
            return INOUT_SEPARATOR.join(item_list)

        # If upstream nodes declared file outputs they need to
        # be downloaded from object storage by the bootstrapper
        if len(cos_inputs) > 0:
            inputs_str = list_to_string(cos_inputs)
            bootstrapper_command.append(f"--inputs '{inputs_str}' ")

        # If this node produces file outputs they need to be uploaded
        # to object storage by the bootstrapper
        if len(cos_outputs) > 0:
            outputs_str = list_to_string(cos_outputs)
            bootstrapper_command.append(f"--outputs '{outputs_str}' ")

        if is_crio_runtime:
            bootstrapper_command.append(f"--user-volume-path '{CRIO_VOL_PYTHON_PATH}' ")

        # Set parameters as env vars; this must be done here rather than adding an
        # env var to the task object because the parameter must be customizable
        if task_parameters:
            parameter_str = list_to_string([f"{param.name}=${param.name}" for param in task_parameters])
            bootstrapper_command.append(
                f"--pipeline-parameters '{parameter_str}' --parameter-pass-method '{self.parameter_pass_method}' "
            )

        command_args.append("".join(bootstrapper_command))

        return command_args

    @staticmethod
    def _sanitize_param_name(name: str) -> str:
        """
        Sanitize a component property or pipeline parameter name.

        Behavior is mirrored from how Kubeflow 1.X sanitizes identifier names:
        - https://github.com/kubeflow/pipelines/blob/1.8.1/sdk/python/kfp/components/_naming.py#L32-L42
        - https://github.com/kubeflow/pipelines/blob/1.8.1/sdk/python/kfp/components/_naming.py#L49-L50
        """
        normalized_name = name.lower()

        # remove non-word characters
        normalized_name = re.sub(r"[\W_]", " ", normalized_name)

        # no double spaces, leading or trailing spaces
        normalized_name = re.sub(" +", " ", normalized_name).strip()

        # no leading digits
        if re.match(r"\d", normalized_name):
            normalized_name = "n" + normalized_name

        return normalized_name.replace(" ", "_")

    def add_disable_node_caching(self, instance: DisableNodeCaching, execution_object: Any, **kwargs) -> None:
        """Add DisableNodeCaching info to the execution object"""
        # Force re-execution of the operation by setting staleness to zero days
        # https://www.kubeflow.org/docs/components/pipelines/overview/caching/#managing-caching-staleness
        if instance.selection:
            execution_object["disable_node_caching"] = True
        else:
            execution_object["disable_node_caching"] = False

    def add_custom_shared_memory_size(self, instance: CustomSharedMemorySize, execution_object: Any, **kwargs) -> None:
        """Add CustomSharedMemorySize info to the execution object"""
        if not instance.size:
            # no custom size was specified; ignore
            return
        execution_object["kubernetes_shared_mem_size"] = {"size": instance.size, "units": instance.units}

    def add_kubernetes_secret(self, instance: KubernetesSecret, execution_object: Any, **kwargs) -> None:
        """Add KubernetesSecret instance to the execution object"""
        if "kubernetes_secrets" not in execution_object:
            execution_object["kubernetes_secrets"] = {}
        execution_object["kubernetes_secrets"][instance.env_var] = {"name": instance.name, "key": instance.key}

    def add_mounted_volume(self, instance: VolumeMount, execution_object: Any, **kwargs) -> None:
        """Add VolumeMount instance to the execution object"""
        if "kubernetes_volumes" not in execution_object:
            execution_object["kubernetes_volumes"] = {}
        execution_object["kubernetes_volumes"][instance.path] = {
            "pvc_name": instance.pvc_name,
            "sub_path": instance.sub_path,
            "read_only": instance.read_only,
        }

    def add_kubernetes_pod_annotation(self, instance: KubernetesAnnotation, execution_object: Any, **kwargs) -> None:
        """Add KubernetesAnnotation instance to the execution object"""
        if "pod_annotations" not in execution_object:
            execution_object["pod_annotations"] = {}
        execution_object["pod_annotations"][instance.key] = instance.value or ""

    def add_kubernetes_pod_label(self, instance: KubernetesLabel, execution_object: Any, **kwargs) -> None:
        """Add KubernetesLabel instance to the execution object"""
        if "pod_labels" not in execution_object:
            execution_object["pod_labels"] = {}
        execution_object["pod_labels"][instance.key] = instance.value or ""

    def add_kubernetes_toleration(self, instance: KubernetesToleration, execution_object: Any, **kwargs) -> None:
        """Add KubernetesToleration instance to the execution object"""
        if "kubernetes_tolerations" not in execution_object:
            execution_object["kubernetes_tolerations"] = {}
        toleration_hash = hashlib.sha256(
            f"{instance.key}::{instance.operator}::{instance.value}::{instance.effect}".encode()
        ).hexdigest()
        execution_object["kubernetes_tolerations"][toleration_hash] = {
            "key": instance.key,
            "operator": instance.operator,
            "value": instance.value,
            "effect": instance.effect,
        }

    @property
    def supported_properties(self) -> Set[str]:
        """A list of Elyra-owned properties supported by this runtime processor."""
        return {
            pipeline_constants.ENV_VARIABLES,
            pipeline_constants.KUBERNETES_SECRETS,
            pipeline_constants.MOUNTED_VOLUMES,
            pipeline_constants.KUBERNETES_POD_ANNOTATIONS,
            pipeline_constants.KUBERNETES_POD_LABELS,
            pipeline_constants.KUBERNETES_TOLERATIONS,
            pipeline_constants.DISABLE_NODE_CACHING,
            pipeline_constants.KUBERNETES_SHARED_MEM_SIZE,
        }

    @property
    def pipeline_parameter_class(self) -> Optional[type]:
        """KfpPipelineProcessor supports KfpPipelineParameter objects."""
        return KfpPipelineParameter


class KfpPipelineProcessorResponse(RuntimePipelineProcessorResponse):
    _type = RuntimeProcessorType.KUBEFLOW_PIPELINES
    _name = "kfp"

    def __init__(self, run_id, run_url, object_storage_url, object_storage_path):
        super().__init__(run_url, object_storage_url, object_storage_path)
        self.run_id = run_id

    def to_json(self):
        response = super().to_json()
        response["run_id"] = self.run_id
        return response



#
# Copyright 2018-2023 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
import os
import shutil
from typing import Any
from typing import List
from typing import Optional
import uuid

from elyra.pipeline.pipeline import GenericOperation
from elyra.pipeline.pipeline import Pipeline
from elyra.pipeline.properties import ElyraProperty


def _read_pipeline_resource(pipeline_filename):
    root = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))
    pipeline_path = os.path.join(root, pipeline_filename)

    with open(pipeline_path, "r") as f:
        pipeline_json = json.load(f)

    return pipeline_json


class NodeBase(object):
    """Represents a node of a constructed pipeline based on files in resources/node_util."""

    id: str
    name: str
    outputs: List[str]
    dependencies: List[str]
    env_vars: List[str]
    image_name: str
    fail: bool
    # from subclasses
    classifier: str
    filename: str
    pipeline_name: str  # Set during pipeline construction

    def __init__(
        self,
        name: str,
        num_outputs: Optional[int] = 0,
        input_nodes: Optional[List[Any]] = None,
        image_name: Optional[str] = None,
        fail: Optional[bool] = False,
    ):
        self.id = str(uuid.uuid4())
        self.name = name
        self.fail = fail
        self.image_name = image_name
        self.outputs = []
        for i in range(1, num_outputs + 1):
            self.outputs.append(f"{self.name}_{i}.out")

        self.inputs = []
        self.parent_operations = []
        if input_nodes:
            for node in input_nodes:
                self.inputs.extend(node.outputs)
                self.parent_operations.append(node.id)

        self.dependencies = ["node_util/*"]

    def get_operation(self) -> GenericOperation:
        env_vars = []
        if self.fail:  # NODE_FILENAME is required, so skip if triggering failure
            if "NODE_FILENAME" in os.environ:  # remove entry if present
                os.environ.pop("NODE_FILENAME")
        else:
            env_vars.append({"env_var": "NODE_FILENAME", "value": self.filename})

        if self.inputs:
            env_vars.append({"env_var": "INPUT_FILENAMES", "value": ";".join(self.inputs)})
        if self.outputs:
            env_vars.append({"env_var": "OUTPUT_FILENAMES", "value": ";".join(self.outputs)})

        # Convey the pipeline name
        assert self.pipeline_name is not None, "Pipeline name has not been set during construction!"
        env_vars.append({"env_var": "PIPELINE_NAME", "value": self.pipeline_name})

        # Add system-owned here with bogus or no value...
        env_vars.append({"env_var": "ELYRA_RUNTIME_ENV", "value": "bogus_runtime"})

        self.env_vars = ElyraProperty.create_instance("env_vars", env_vars)

        component_parameters = {
            "filename": self.filename,
            "runtime_image": self.image_name or "NA",
            "dependencies": self.dependencies,
            "inputs": self.inputs,
            "outputs": self.outputs,
        }
        return GenericOperation(
            self.id,
            "execution_node",
            self.name,
            self.classifier,
            parent_operation_ids=self.parent_operations,
            component_props=component_parameters,
            elyra_props={"env_vars": self.env_vars},
        )


class NotebookNode(NodeBase):
    def __init__(
        self,
        name: str,
        num_outputs: Optional[int] = 0,
        input_nodes: Optional[List[Any]] = None,
        image_name: Optional[str] = None,
        fail: Optional[bool] = False,
    ):
        super().__init__(name, num_outputs=num_outputs, input_nodes=input_nodes, image_name=image_name, fail=fail)
        self.classifier = "execute-notebook-node"
        self.filename = f"{self.name}.ipynb"


class PythonNode(NodeBase):
    def __init__(
        self,
        name: str,
        num_outputs: Optional[int] = 0,
        input_nodes: Optional[List[Any]] = None,
        image_name: Optional[str] = None,
        fail: Optional[bool] = False,
    ):
        super().__init__(name, num_outputs=num_outputs, input_nodes=input_nodes, image_name=image_name, fail=fail)
        self.classifier = "execute-python-node"
        self.filename = f"{self.name}.py"


def construct_pipeline(
    name: str,
    nodes: List[NodeBase],
    location: str,
    runtime_type: Optional[str] = "local",
    runtime_config: Optional[str] = "local",
) -> Pipeline:
    """Returns an instance of a local Pipeline consisting of each node and populates the
    specified location with the necessary files to run the pipeline from that location.
    """
    pipeline = Pipeline(str(uuid.uuid4()), name, runtime_type, runtime_config)
    for node in nodes:
        node.pipeline_name = name
        pipeline.operations[node.id] = node.get_operation()
        # copy the node file into the "working directory"
        if isinstance(node, NotebookNode):
            src_file = os.path.join(os.path.dirname(__file__), "resources/node_util/node.ipynb")
        elif isinstance(node, PythonNode):
            src_file = os.path.join(os.path.dirname(__file__), "resources/node_util/node.py")
        else:
            assert False, f"Invalid node type detected: {node.__class__.__name__}"

        shutil.copy(src_file, os.path.join(location, node.filename))

    # copy the node_util directory into the "working directory"
    shutil.copytree(os.path.join(os.path.dirname(__file__), "resources/node_util"), os.path.join(location, "node_util"))

    return pipeline
